#!/usr/bin/env perl
#
#  program:	charmm.pl
#  author:	Pieter J. in 't Veld
#  date:	April 30 - May 5, June 21-26, 2013, January 1, 2015,
#  		January 9, 15, 19, August 4, October 14, 2016, March 3,
#  		December 12, 2017, February 14, 2018, October 27, 2019,
#  		March 20-31, May 6, June 25, 2020
#  purpose:	Interpret CHARMM force field files and convert them into
#  		EMC textual force field formats; part of EMC distribution
#
#  Copyright (c) 2004-2022 Pieter J. in 't Veld
#  Distributed under GNU Public License as stated in LICENSE file in EMC root
#  directory
#
#  notes:
#    20130430	Conception date; based on opls.pl
#    20130804	Added inner and cutoff definitions to define
#    20131106	Added ring type identification to define
#    20150131	Added inclusion of .inp files
#    20150218	Cleaned up connectivity for rules of rings with c or n
#    20160109	Added mass recognition
#    		Streamlined partial charge interpretation
#    20160115	Added aromatic recognition for CGenFF
#    		Improved topology file interpretation
#    20160119	Allowed for multiple connections in SMILES to a single atom
#    20160123	Added shake flag
#    20160804	Added ring exceptions for CGENFF
#    20161014	Adapted sulphur and nitrogen bond interpretations
#    20170311	Added FFAPPLY force field application style (either 'all' or 
#		'groups')
#    20171210	Added FFINDEX for how to order rules (either 'types' or
#    		'elements')
#    		Cleaned up mass interpretation
#    20180214	Corrected patch interpretation
#    20191027	Added matrix and SVD solutions for bond increment determination
#    20200320	Cleaned up bond increment determination
#    20200323	Added item average to average charges of specified atom IDs
#    		Added item combine to combine all bonded contributions of
#    		specified type IDs
#    20200324	Added item include for defining only included residues
#    20200328	Generalized parameter printing routines
#    20200331	Overhaul of programmatic flow to promote automated retyping
#    		for increments
#    20200506	Added read increments to field increment list
#    20200625	Improved internal rule sorting
#    20200725	Improved force field file recognition
#    20200815	Improved patch creation
#    20210401	Included .inp files
#

use strict;

use Data::Dumper;
use File::Basename;
use Fcntl qw(:seek); 

BEGIN {
  use File::Basename;
  if (-d dirname($0)."/modules") {
    $::EMCCharmm::Modules{dir} = dirname($0)."/modules";
  } else {
    $::EMCCharmm::Modules{dir} = $ENV{EMC_ROOT}."/scripts/modules";
  }
  if (! -d $::EMCCharmm::Modules{dir}) {
    print("Error: cannot locate modules directory: set EMC_ROOT?\n\n");
    exit(-1);
  }
}
use lib $::EMCCharmm::Modules{dir};
use EMC::Matrix;
use EMC::Message;


# functions

$::EMCCharmm::script = "emc_charmm.pl";
$::EMCCharmm::author = "Pieter J. in 't Veld";
$::EMCCharmm::version = "v3.3";
$::EMCCharmm::date = "April 1, 2021";
$::EMCCharmm::EMCVersion = "9.4.4";
$::EMCCharmm::TOPOL = 0;
$::EMCCharmm::PARMS = 0;
$::EMCCharmm::DEFINE = 0;
$::EMCCharmm::Round = 1e-5;
$::EMCCharmm::SVDRound = 1e-3;
%::EMCCharmm::Type = ();
%::EMCCharmm::Index = ();
%::EMCCharmm::Masses = ();
$::EMCCharmm::MassOrder = 0;
%::EMCCharmm::Element = ("*" => "*");
%::EMCCharmm::Nonbond = ();
%::EMCCharmm::Bond = ();
%::EMCCharmm::BondAuto = ();
%::EMCCharmm::Angle = ();
%::EMCCharmm::AngleAuto = ();
%::EMCCharmm::Torsion = ();
%::EMCCharmm::TorsionAuto = ();
%::EMCCharmm::Improper = ();
%::EMCCharmm::ImproperAuto = ();
%::EMCCharmm::Equivalence = ();
%::EMCCharmm::EquivalenceAuto = ();
@::EMCCharmm::References = ();
@::EMCCharmm::Precedences = ();
%::EMCCharmm::Increments = ();
%::EMCCharmm::Templates = ();
%::EMCCharmm::Replace = ();
%::EMCCharmm::Rules = ();
%::EMCCharmm::Extras = ();
$::EMCCharmm::Dummy = "DUM";
$::EMCCharmm::Wildcard = "*";
$::EMCCharmm::Anything = "?";
$::EMCCharmm::FFMode = "CHARMM";
$::EMCCharmm::FFApply = "GROUPS";
$::EMCCharmm::FFIndex = "TYPES";
$::EMCCharmm::FFRule = "ERROR";
$::EMCCharmm::FFVersion = "UNKNOWN";
$::EMCCharmm::FFDepth = 0;
$::EMCCharmm::FFDepthRead = 0;
$::EMCCharmm::FFDepthMax = 5;
$::EMCCharmm::FFCharge = "ALL";
$::EMCCharmm::FFShake = "NONE";
$::EMCCharmm::Cutoff = 12.0;
$::EMCCharmm::Inner = 10.0;
$::EMCCharmm::Date = localtime();

%::EMCCharmm::Allowed = (
  apply => {
    all => "all", none => "all", general => "all",
    groups => "groups", template => "groups"},
  index => {
    default => "types", types => "types", elements => "elements"},
  rule => {
    default => "error", ignore => "ignore", error => "error"
  }
);

%::EMCCharmm::Flag = (
  debug => 0,
  increment => 0,
  mass => 0,
  message => 1,
  reduce => 0,
  show => 0,
  sort => 1,
  warning => 1
);

%::EMCCharmm::Journals = (
  "Biophysical Journal"				=>	"Biophys. J.",
  "Journal de Chimie Physique"			=>	"J. Chim. Phys.",
  "Journal of the American Chemical Society"	=>	"J. Am. Chem. Soc.",
  "Journal of Computational Chemistry"		=>	"J. Comput. Chem.",
  "Journal of Chemical Physics"			=>	"J. Chem. Phys.",
  "Journal of Physical Chemistry B"		=>	"J. Chem. Phys. B",
  "PROTEINS: Structure, Function, and Genetics"	=>	"Proteins: Struct., Funct., Genet."
);

%::EMCCharmm::Residue = (
  charge => {},
  exclude => {},
  excludes => [],
  include => {},
  includes => [],
  index => {},
  patch => {},
  patches => [],
  residues => [],
  type => {},
  wildcard => {},
  lists => ["excludes", "includes", "patches", "residues"]
);


# initial settings

sub flag {
  return @_[0] eq ""		? 1 : 
    lc(@_[0]) eq "true"		? 1 : 
    lc(@_[0]) eq "false"	? 0 : eval(@_[0]);
}


sub boolean {
  return @_[0] if (@_[0] eq "true");
  return @_[0] if (@_[0] eq "false");
  return @_[0] ? @_[0]<0 ? "auto" : "true" : "false";
}


sub check_shake {
  foreach ("NONE", "HYDROGEN", "WATER", "ALL") { return if ($_ eq uc(@_[0])); }
  error("unknown shake mode '@_[0]'");
}


sub header {
  message(
    "CHARMM field conversion $::EMCCharmm::version, ".
    "$::EMCCharmm::date, (c) $::EMCCharmm::author\n\n");
}


sub round {
  my $round = @_[1] ne "" ? @_[1] :
	      $::EMCCharmm::Round>0 ? $::EMCCharmm::Round : 0.001;
  return int(@_[0]/$round+(@_[0]<0 ? -1 : 1)*(0.5+1e-10))*$round;
}


sub list_round {
  my @result;

  foreach (@_) {
    push(@result, round($_));
  }
  return @result;
}


sub trim {
  my $s = shift(@_);
  $s =~ s/^\s+|\s+$//g;
  return $s;
}


sub help {
  $::EMCCharmm::Flag{message} = 1;
  header();
  message("Usage:\n  $::EMCCharmm::script [-option[=value]] project [additional [..]]\n");
  message("\nOptions:\n");
  message("  -define\tSet alternative define file name\n");
  message("  -depth\tSet maximum rule recursion depth [$::EMCCharmm::FFDepthMax]\n");
  message("  -doubles\tShow occurence of double entries\n");
  message("  -exclude\tSet to be excluded residue\n");
  message("  -info\t\tTurn on messages and info\n");
  message("  -include\tSet to be included residue\n");
  message("  -increments\tTurn on determination of bond increments [".boolean($::EMCCharmm::Flag{increment})."]\n");
  message("  -literature\tShow literature references\n");
  message("  -output\tSet alternative output field name\n");
  message("  -quiet\tTurn off all output\n");
  message("  -reduce\tReduce number of types\n");
  message("  -shake\tSet shake mode [$::EMCCharmm::FFShake]\n");
  message("  -show\t\tShow intermediate increment matrices [".boolean($::EMCCharmm::Flag{show})."]\n");
  message("  -source\tProvide alternate source directory\n");
  message("  -trial\tDebug a specific residue [$::EMCCharmm::FFTrial]\n");
  message("  -apply\tSet force field application style [$::EMCCharmm::FFApply]\n");
  message("  -rule\t\tSet force field rule interpretation style [$::EMCCharmm::FFIndex]\n");
  message("\nNotes:\n");
  message("  * Assumes existing input files: par_project.prm, top_project.rtf, and\n");
  message("    project.define\n");
  message("  * Optional additional force field files par_additional.inp and\n");
  message("    top_additional.inp are concatenated with chosen force field\n");
  message("  * Allowed styles are 'general' or 'template', respectively \n");
  message("    applying to either all groups or groups specified under the\n");
  message("    'TEMPLATES' paragraph\n");
  message("  * Included and excluded residues are accumulative\n");
  message("  * All topology residues are considered when include is not set\n");
  message("  * Creates output: project.prm and project.top\n");
  message("\n");
  exit;
}


sub init {
  my %check;
  my @names = ();
  my $name = "";
  my %ext = (rtf => 1, prm => 2, str => 3);
  my @pre = ("", "top", "par", "toppar");

  $::EMCCharmm::DefineName = "";
  $::EMCCharmm::ParameterName = "";
  foreach (@_) {
    if (substr($_, 0, 1) eq "-") {
      my @arg = split("=");
      if ($arg[0] eq "-apply") {
       	if (!defined(${$::EMCCharmm::Allowed{apply}}{lc($arg[1])})) {
	  error("illegal apply style '$arg[1]'\n");
       	}
	$::EMCCharmm::FFApply = uc(${$::EMCCharmm::Allowed{apply}}{lc($arg[1])});
      }
      elsif ($arg[0] eq "-debug") { 
	$::EMCCharmm::Flag{debug} = flag(@arg[1]);
      }
      elsif ($arg[0] eq "-define") {
       	$::EMCCharmm::DefineName = strip($arg[1], ".define");
      }
      elsif ($arg[0] eq "-depth") {
       	$::EMCCharmm::FFDepthMax = eval($arg[1]);
      }
      elsif ($arg[0] eq "-doubles") { 
	$::EMCCharmm::Doubles = 1;
      }
      elsif ($arg[0] eq "-frequency") {
      }
      elsif ($arg[0] eq "-exclude") { 
	$::EMCCharmm::Residue{exclude}->{trim($arg[1])} = 1;
      }
      elsif ($arg[0] eq "-include") { 
	$::EMCCharmm::Residue{include}->{trim($arg[1])} = 1;
      }
      elsif ($arg[0] eq "-increments") { 
	$::EMCCharmm::Flag{increment} = flag(@arg[1]);
      }
      elsif ($arg[0] eq "-index") {
       	if (!defined(${$::EMCCharmm::Allowed{index}}{lc($arg[1])})) {
	  error("illegal index style '$arg[1]'\n");
	}
	$::EMCCharmm::FFIndex = uc(${$::EMCCharmm::Allowed{index}}{lc($arg[1])});
      }
      elsif ($arg[0] eq "-info") { 
	$::EMCCharmm::Flag{message} = 1;
      }
      elsif ($arg[0] eq "-mass") {
	$::EMCCharmm::Flag{mass} = flag($arg[1]);
      }
      elsif ($arg[0] eq "-literature") {
       	$::EMCCharmm::Literature = 1;
      }
      elsif ($arg[0] eq "-output") {
       	$name = $arg[1];
      }
      elsif ($arg[0] eq "-quiet") { 
	$::EMCCharmm::Flag{message} = 0;
      }
      elsif ($arg[0] eq "-rule") {
       	if (!defined(${$::EMCCharmm::Allowed{rule}}{lc($arg[1])})) {
	  error("illegal rule style '$arg[1]'\n");
	}
	$::EMCCharmm::FFRule = uc(${$::EMCCharmm::Allowed{rule}}{lc($arg[1])});
      }
      elsif ($arg[0] eq "-reduce") {
       	$::EMCCharmm::Flag{reduce} = flag($arg[1]);
      }
      elsif ($arg[0] eq "-shake") {
       	$::EMCCharmm::FFShake = uc($arg[1]);
      }
      elsif ($arg[0] eq "-show") {
       	$::EMCCharmm::Flag{show} = flag($arg[1]);
      }
      elsif ($arg[0] eq "-source") { 
	$::EMCCharmm::SourceDir = $arg[1];
      }
      elsif ($arg[0] eq "-trial") {
       	$::EMCCharmm::FFTrial = $arg[1];
      }
      else { help(); }
      next;
    }
    my $flag = 1;						# scrub name
    my ($name, $path, $suffix) = fileparse($_, '\.[^\.]*');
    if (defined($ext{$suffix})) {
      $flag = $ext{$suffix}; $name =~ s/^($pre[$flag])\s+//; }
    next if ($check{$_});
    $check{$_} = $flag ;
    push(@names, $_);
  }
  help() if (scalar(@names)<1);
  $name = $names[0] if ($name eq "");
  my @arg = split("/", $::EMCCharmm::SourceDir);
  $::EMCCharmm::SourceDir = scalar(@arg) ? join("/", @arg)."/" : "";
  $::EMCCharmm::DefineName = ($::EMCCharmm::DefineName eq "" ? @names[0] : $::EMCCharmm::DefineName).".define";
  $::EMCCharmm::CreatedParameterName = $name.".prm";
  $::EMCCharmm::CreatedTopologyName = $name.".top";
  $::EMCCharmm::ProjectName = $name;
  header();
  check_shake($::EMCCharmm::FFShake);
  files(@names);
}


sub strip {
  return (split(@_[1], @_[0]))[0];
}


sub spool {
  my $stream = shift(@_);
  my $input = fopen(@_[0], "r");

  foreach (<$input>) {
    chop;
    if ($_ eq "THETAS") { $_ = "\nANGLES"; }
    elsif ($_ eq "PHI") { $_ = "\nDIHEDRALS"; }
    printf($stream "%s\n", $_);
  }
  fclose($input);
}


@::EMCCharmm::ParameterIndex = ();
@::EMCCharmm::TopologyIndex = ();
@::EMCCharmm::StreamIndex = ();

sub files {
  my @names = @_;
  my $stream = fopen(
    $::EMCCharmm::ParameterName = 
    $::EMCCharmm::TopologyName = "toppar_".$$.".str", "w");

  foreach(@names) {
    my $found = 0;
    my @arg = split("/");
    my $name = pop(@arg);
    my $dir = scalar(@arg) ? join("/", @arg)."/" : "";
    if ($::EMCCharmm::SourceDir ne "") {
      $dir = $::EMCCharmm::SourceDir.($dir eq "" ? "" : $dir);
    }
    if (-f $dir."par_".$name.".prm") {
      push(@::EMCCharmm::ParameterIndex, tell($stream));
      spool($stream, $dir."par_".$name.".prm");
      $found += 1;
    }
    elsif (-f $dir."par_".$name.".inp") {
      push(@::EMCCharmm::ParameterIndex, tell($stream));
      spool($stream, $dir."par_".$name.".prm");
      $found += 1;
    }
    elsif (-f $dir.$name.".prm") {
      push(@::EMCCharmm::ParameterIndex, tell($stream));
      spool($stream, $dir.$name.".prm");
      $found += 1;
    }
    if (-f $dir."top_".$name.".rtf") {
      push(@::EMCCharmm::TopologyIndex, tell($stream));
      spool($stream, $dir."top_".$name.".rtf");
      $found += 1;
    }
    elsif (-f $dir."top_".$name.".inp") {
      push(@::EMCCharmm::TopologyIndex, tell($stream));
      spool($stream, $dir."top_".$name.".rtf");
      $found += 1;
    }
    elsif (-f $dir.$name.".rtf") {
      push(@::EMCCharmm::TopologyIndex, tell($stream));
      spool($stream, $dir.$name.".rtf");
      $found += 1;
    }
    if (-f $dir."toppar_".$name.".str") {
      push(@::EMCCharmm::StreamIndex, tell($stream));
      spool($stream, $dir."toppar_".$name.".str");
      $found += 2;
    }
    elsif (-f $dir.$name.".str") {
      push(@::EMCCharmm::StreamIndex, tell($stream));
      spool($stream, $dir.$name.".str");
      $found += 2;
    }
    if ($found!=2) {
      error("file not found for '$name'");
    }
  }
  close($stream);
}


# topology corrections due to deficiencies in topology file

sub hash_cmp {
  my @a = split("\t", $a);
  my @b = split("\t", $b);
  my @s = (scalar(@a), scalar(@b));

  return -1 if (@s[1]<@s[0]);
  return 1 if (@s[1]>@s[0]);
  
  for (my $i=0; $i<@s[0]; ++$i)
  {
    my @l = (length(@a[$i]), length(@b[$i]));
    return -1 if (@l[1]<@l[0]);
    return 1 if (@l[1]>@l[0]);
  }
  return 0;
}


sub hash_entry {
  my $hash = shift(@_);

  foreach (sort(hash_cmp keys(%{$hash}))) {
    my $i = -1;
    my @key1 = @_; 
    my @key2 = reverse(@key1);
    
    foreach (split("\t")) {
      my $index = index($_, "*");
      ++$i; next if ($index<0);
      @key1[$i] = substr(@key1[$i], 0, $index)."*";
      @key2[$i] = substr(@key2[$i], 0, $index)."*";
    }
    return defined($hash->{$_}) ? $hash->{$_} : undef if ((
      $_ eq join("\t", arrange(@key1)) ||
      $_ eq join("\t", arrange(@key2)))
    );
  }
  return undef;
}


sub ring {
  return 0 if (substr(@_[0],0,1) eq "H");
  my $entry = hash_entry(\%::EMCCharmm::Ring, @_);
  return $entry->{n} if (defined($entry));
  return 0;
}


sub aromatic {
  my $entry = hash_entry(\%::EMCCharmm::Ring, @_);
  return $entry->{aromatic} if (defined($entry));
  return 0;
}


sub repair_bond {
  my $residue = shift(@_);
  my $atom = shift(@_);
  my $bond = shift(@_);
  my $type = shift(@_);
  my $connect = shift(@_);
  my $nconnects = shift(@_);
  my %visited = ();
  my %double = ();
  my $flag = 0;

  foreach (@{$atom}) {
    my $a = $_;
    foreach (@{$connect->{$_}}) {
      my $key = join("\t", arrange($a, $_));
      next if (defined($visited{$key}));
      $visited{$key} = 1;
      my @t = arrange($type->{$a}, $type->{$_});
      my $order = hash_entry(\%::EMCCharmm::BondOrder, @t);
      next if (!defined($order));
      my @e = arrange(
	lc($::EMCCharmm::Element{@t[0]}), lc($::EMCCharmm::Element{@t[1]}));
      my $tmp = $order;
      if ($order==2 && @e[0] eq "c" && @e[1] eq "c") {
	$order = 1 if (defined($double{$a}) || defined($double{$_}));
	$double{$a} = $double{$_} = 1 if ($order==2);
      }
      $bond->{$key} = $order;
      #tprint(__LINE__.":", "REPAIR", $key, @t, @e, $tmp, $order);
    }
  }

  foreach (keys(%{$bond})) {
    my $key = $_;
    my @arg = split("\t");
    my @t = arrange($type->{@arg[0]}, $type->{@arg[1]});
    my @s = (substr(@t[0],0,2), substr(@t[1],0,2));
    my @r = (substr(@t[0],0,4), substr(@t[1],0,4));

    @arg[0,1] = @arg[1,0] if (join("\t", @s) ne join("\t", @t));
    if ((aromatic(@s[0], $residue) && 
	 aromatic(@s[1], $residue)) ||
	(aromatic($type->{@arg[0]}, $residue) && 
	 aromatic($type->{@arg[1]}, $residue))) {
      $bond->{$_} = 1.5;
    } elsif (@arg[0] eq "O2" || @arg[1] eq "O2") {
      $flag = 1;
    }
  }
  
  return if ($flag);

  foreach (keys(%{$type})) {
    next if (!($_ eq "O2" && $type->{$_} eq "OCL"));
    $bond->{"C1\tO2"} = 1.5 if (!defined($bond->{"C1\tO2"}));
  }
}


# general functions

sub error {
  printf("Error: ".shift(@_)."%s\n", join(" ", @_));
  exit(-1);
}


sub error_line {
  my $line = shift(@_);
  if ($line<0) { error(@_); }
  else { 
    my $format = shift(@_);
    $format =~ s/\n/ in line $line of input\n/g;
    if (scalar(@_)) { error($format, @_); }
    else { error($format); }
  }
}


sub debug {
  printf("Debug: ".shift(@_)."%s", join(" ", @_)) if ($::EMCCharmm::Flag{debug});
}


sub tdebug {
  print(join("\t", "Debug:", @_), "\n") if ($::EMCCharmm::Flag{debug});
}


sub info {
  printf("Info: ".shift(@_)."%s", join(" ", @_)) if ($::EMCCharmm::Flag{message});
}


sub warning {
  printf("Warning: ".shift(@_)."%s", join(" ", @_)) if ($::EMCCharmm::Flag{warning});
}


sub message {
  printf(shift(@_)."%s", join(" ", @_)) if ($::EMCCharmm::Flag{message});
}


sub tprint {
  print(join("\t", @_), "\n");
}


sub ftest {
  my $name = shift(@_);
  my ($result, $file);

  open($file, "<$name");
  $result = scalar($file) ? 1 : 0;
  close($file);
  return $result;
}


sub fopen {
  my $name = shift(@_);
  my $mode = shift(@_);
  my $ignore = shift(@_);
  my ($file, $result);

  return 0 if ($name eq "");
  error("illegal mode") if (!($mode eq "r" || $mode eq "w"));
  info("opening \"$name\" for ", ($mode eq "r" ? "read" : "writ")."ing\n");
  open($file, ($mode eq "r" ? "<" : ">").$name);
  if (!scalar(stat($file))) {
    error("cannot open file \"$name\"") if ($ignore);
    warning("cannot open file \"$name\"\n");
    $file = 0;
  }
  return $file;
}


sub fclose {
  my $file = shift(@_);

  close($file) if (scalar($file));
  return 0;
}


sub open_all {
  if (!scalar($::EMCCharmm::TOPOL)) {
    $::EMCCharmm::TOPOL = fopen($::EMCCharmm::TopologyName, "r");
  }
  if (!scalar($::EMCCharmm::PARMS)) {
    $::EMCCharmm::PARMS = fopen($::EMCCharmm::ParameterName, "r");
  }
  if (!scalar($::EMCCharmm::DEFINE)) {
    $::EMCCharmm::DEFINE = fopen($::EMCCharmm::DefineName, "r");
  }

  return if (defined($::EMCCharmm::FFTrial));
  
  if (!scalar($::EMCCharmm::OUT_PARMS)) {
    $::EMCCharmm::OUT_PARMS = fopen($::EMCCharmm::CreatedParameterName, "w");
  }
  if (!scalar($::EMCCharmm::OUT_TOPOL)) {
    $::EMCCharmm::OUT_TOPOL = fopen($::EMCCharmm::CreatedTopologyName, "w");
  }
}


sub close_all {
  $::EMCCharmm::TOPOL = fclose($::EMCCharmm::TOPOL);
  $::EMCCharmm::PARMS = fclose($::EMCCharmm::PARMS);
  $::EMCCharmm::DEFINE = fclose($::EMCCharmm::DEFINE);
  $::EMCCharmm::OUT_PARMS = fclose($::EMCCharmm::OUT_PARMS);
  $::EMCCharmm::OUT_TOPOL = fclose($::EMCCharmm::OUT_TOPOL);

  unlink $::EMCCharmm::ParameterName;
  unlink $::EMCCharmm::TopologyName;
}


# bonded arrangements

sub arrange_compare {
  my ($a, $b) = @_[0,1];
  return 0 if ($a eq "*");
  return 1 if ($b eq "*");
  $a = substr($a,1) if (substr($a,0,1) =~ m/\+|\-]/);
  $b = substr($b,1) if (substr($b,0,1) =~ m/\+|\-]/);
  return $a lt $b ? 1 : 0;
}


sub arrange {
  return @_ if (scalar(@_)<2);
  return @_ if (arrange_compare(@_[0, -1]));
  if (@_[0] eq @_[-1]) {
    return @_ if (scalar(@_)<4);
    return @_ if (arrange_compare(@_[1, 2]));
    return @_[3,2,1,0];
  }
  return @_[1,0] if (scalar(@_) == 2);
  return @_[2,1,0] if (scalar(@_) == 3);
  return @_[3,2,1,0];
}


sub arrange_imp {
#  return @_[0,2,3,1] if (
#    (arrange_compare(@_[2,3])||(@_[2] eq @_[3])) && arrange_compare(@_[2, 1]));
#  return @_[0,3,1,2] if (
#    (arrange_compare(@_[3,1])||(@_[3] eq @_[1])) && arrange_compare(@_[3, 2]));
  return @_;
}


sub redefine {
  return defined($::EMCCharmm::Redefine{@_[0]}) ? $::EMCCharmm::Redefine{@_[0]} : @_[0];
}


# handling of define file

sub split_data {
  my @arg = split("\t", @_[0]);
  my @result = split(",", @arg[0]);

  @result = scalar(split(":", @result[0]))>1 ? (shift(@arg)) : ();
  foreach(@arg) { 
    foreach (split(",", $_)) { 
      push(@result, $_);
    }
  }
  @result = split(" ", @_[0]) if (scalar(@result)==1);
  foreach (@result) { 
    $_ = trim($_);
  }
  @arg = split(" ", @result[0]);
  if (scalar(@arg)>1) {
    shift(@result);
    unshift(@result, @arg);
  }
  push (@result, ",") if (substr(@_[0],-1,1) eq ",");
  @arg = (); 
  foreach (@result) {
    last if (substr($_,0,1) eq "#");
    push(@arg, $_) if ($_ ne "");
  }
  return @arg;
}


sub get_define {
  return if (!scalar($::EMCCharmm::DEFINE));

  my $read = 0;
  my %item = (
    end => 0, define => 1, mass => 2, precedence => 3, equivalence => 4, 
    extra => 5, bond => 6, angle => 7, torsion => 8, improper => 9, 
    references => 10, rules => 11, redefine => 12, repair => 13, replace => 14,
    patch => 15, wildcard => 16, exclude => 17, delete => 18, ring => 19,
    nonbond => 20, charge => 21, order => 22, uncouple => 23, increment => 24,
    retype => 25, comment => 26, average => 27, combine => 28, include => 29,
    replica => 30, last => 99
  );
  my @command = (); foreach (keys(%item)) { @command[$item{$_}] = $1; }
  my @message = ("", "defines", "extra masses", "precedences", "equivalences",
    "extra types", "extra bonds", "extra angles", "extra torsions",
    "extra impropers", "references", "rules", "type redefinitions", "repairs",
    "rule replacements", "residue patches", "residue wildcard",
    "excluded residues", "rule deletions", "ring types", "extra nonbonds",
    "charge redefinitions", "bond order", "bond increment uncoupling",
    "bond increments", "retypes", "comments", "averages", "combinations",
    "included residues", "replicas", ""
  );
  my @nargs = (
    1, 1, 2, 1, 3, 
    3, 4, 5, 7, 
    6, 4, 3, 2, 3,
    5, 1, 1,
    1, 0, 1, 6,
    1, 3, 2,
    3, 3, 0, 3, 3,
    1
  );
  my @command_stack = ();
  my %command_level = ();
  my @command_extra = ();
  my @command_flag = ();
  my $comment = 0;
  my $line = 0;
  my $rank = 0;
  my $command;

  seek($::EMCCharmm::DEFINE, 0, 0);
  while (<$::EMCCharmm::DEFINE>) {
    chop(); ++$line;
    next if (substr($_, 0, 1) eq "#");
    my @arg = split_data($_);
    if (@arg[0] eq "ITEM") {					# find command
      if (lc(@arg[1]) eq "last") {
	next if ($command_level{comment});
	last;
      } elsif (lc(@arg[1]) eq "end") {
	my $flag = pop(@command_flag);
	$command = pop(@command_stack);
	if (
	  $flag ? $command eq "" || !$command_level{$command} : 0) {
	  error_line($line, "unmatched end\n");
	}
	--$command_level{$command} if ($flag);
	$command = "end";
      } else {
	my @a = split("=", @arg[2]);
	my $f = substr(@a[1],0,1) eq "!" ? 1 : 0;
	   @a[1] = substr(@a[1],1) if ($f);
	my $flag = lc(@a[0]) eq "flag" || scalar(@a)==1 ? lc(@a[1]) : "";
	$command = lc(@arg[1]);
	@command_extra = $flag eq "" ? @arg[2..$#arg] : ();
	$flag = $flag eq "" ? 1 : $::EMCCharmm::Flag{$flag} ? $f^1 : $f;
	++$command_level{$command} if ($flag);
	push(@command_stack, $command);
	push(@command_flag, $flag);
      }
      if (!defined($item{$command})) {
	error_line($line, "undefined item '$command'\n");
      }
      $read = $item{$command};
      if ($read && !$command_level{comment}) {
	info("reading ".@message[$read]."\n");
      }
      next;
    }
    next if (!$read);
    next if ($command_level{comment});
    next if (!scalar(@arg));
    next if ((@nargs[$read]>0)&&(scalar(@arg)<@nargs[$read]));
    if ($read==1) {						# define
      my @ARG = @arg; foreach (@ARG) { $_ = uc($_); }
      if ($ARG[0] eq "FFMODE") { $::EMCCharmm::FFMode = $ARG[1]; } 
      elsif ($ARG[0] eq "FFAPPLY") {
       	if (!defined(${$::EMCCharmm::Allowed{apply}}{lc($ARG[1])})) {
	  error_line($line, "illegal apply style '$ARG[1]'\n");
       	}
	$::EMCCharmm::FFApply = uc(${$::EMCCharmm::Allowed{apply}}{lc($ARG[1])});
      } 
      elsif ($ARG[0] eq "FFCHARS") { 
	$::EMCCharmm::TypeNChars = eval($ARG[1]);
      }
      elsif ($ARG[0] eq "FFDEPTH") { $::EMCCharmm::FFDepthRead = $ARG[1]; } 
      elsif ($ARG[0] eq "FFINCREMENT") {
	$::EMCCharmm::Flag{increment} = flag($ARG[1]); } 
      elsif ($ARG[0] eq "FFINDEX") {
       	if (!defined(${$::EMCCharmm::Allowed{index}}{lc($ARG[1])})) {
	  error_line($line, "illegal index style '$ARG[1]'\n");
       	}
	$::EMCCharmm::FFIndex = uc(${$::EMCCharmm::Allowed{index}}{lc($ARG[1])});
      }
      elsif ($ARG[0] eq "FFMASS") { $::EMCCharmm::Flag{mass} = flag($ARG[1]); }
      elsif ($ARG[0] eq "FFROUND") { $::EMCCharmm::Round = eval($ARG[1]); }
      elsif ($ARG[0] eq "FFSORT") { $::EMCCharmm::Flag{sort} = flag($ARG[1]); }
      elsif ($ARG[0] eq "FFSOURCE") { $::EMCCharmm::FFSource = $ARG[1]; }
      elsif ($ARG[0] eq "VERSION") {
	shift(@arg); $::EMCCharmm::FFVersion = join(" ", @arg); } 
      elsif ($ARG[0] eq "CUTOFF") { $::EMCCharmm::Cutoff = $ARG[1]; } 
      elsif ($ARG[0] eq "INNER") { $::EMCCharmm::Inner = $ARG[1]; } 
      elsif ($ARG[0] eq "SHAKE") { 
	check_shake($::EMCCharmm::FFShake = $ARG[1]); }
      elsif ($ARG[0] eq "RULE") { 
       	if (!defined(${$::EMCCharmm::Allowed{rule}}{lc($ARG[1])})) {
	  error_line($line, "illegal rule style '$ARG[1]'\n");
       	}
	$::EMCCharmm::FFRule = uc(${$::EMCCharmm::Allowed{rule}}{lc($ARG[1])});
      } 
    } elsif ($read==2) {					# masses
      my $dest = shift(@arg);
      my $src = shift(@arg);
      unshift(@arg, $dest);
      @arg = cpy_masses($src, @arg) if ($src ne "-");
      set_masses(@arg[6,4,0,1,2,5]);
    } elsif ($read==30) {					# replica
      if (scalar(@arg)!=2) {
	error_line($line, "incorrect number of arguments for replica\n");
      }
      if (defined($::EMCCharmm::Replica{@arg[0]})) {
	error_line($line, "replica '@arg[0]' already defined\n");
      }
      $::EMCCharmm::Replica{@arg[0]} = {src => @arg[1]};
    } elsif ($read==3) {					# precedence
      next if ($_ eq "");
      push(@::EMCCharmm::Precedences, $_);
    } elsif ($read==4) {					# equivalence
      $::EMCCharmm::Type{$arg[1]} = 1;
      while (scalar(@arg)<7) { push(@arg, @arg[-1]); }
      my $id = shift(@arg);
      $::EMCCharmm::Equivalence{$id} = [@arg];
      $::EMCCharmm::EquivalenceAuto{$id} = [@arg];
    } elsif ($read==10) {					# references
      push(@::EMCCharmm::References, join("\t", @arg[0,3,1,2]));
    } elsif ($read==11) {					# rules
      #@arg[1] = sprintf("%.10g", @arg[1]);
      #$arg[-1] =~ s/\^//g;
      my $index = join("\t", @arg[-2,-1]);
      if (scalar(@arg)==5) {
	$::EMCCharmm::Rules{$index} = join("\t", @arg[0,1,2]);
      } else {
	$::EMCCharmm::Rules{$index} = join("\t", @arg[0], "-", "-");
      }
    } elsif ($read==14) {					# replace
      #@arg[3] = sprintf("%.10g", @arg[3]);
      #$arg[-1] =~ s/\^//g;
      my $index = join("\t", @arg[0,1,2]);
      $::EMCCharmm::Replace{$index} = join("\t", @arg[3,4]);
    } elsif ($read==12) {					# redefine
      $::EMCCharmm::Redefine{$arg[0]} = $arg[1];
    } elsif ($read==21) {					# charge
      if (scalar(@command_extra)) {
	foreach (@command_extra) {
	  $::EMCCharmm::Charge{join("\t", $_, @arg[1])} = $arg[2];
	}
      } else {
	$::EMCCharmm::Charge{join("\t", @arg[0,1])} = $arg[2];
      }
    } elsif ($read==13) {					# repair
      my %n = (bond => 2);
      if (!defined($n{@arg[0] = lc(@arg[0])})) {
	error_line($line, "unallowed keyword '@arg[0]'\n");
      }
      if (!((scalar(@arg)==2+$n{@arg[0]})||(scalar(@arg)!=2+2*$n{@arg[0]}))) {
	error_line($line, "incorrect number of arguments\n");
      }
      if (@arg[0] eq "bond") {
	my $key = join("\t", arrange(@arg[2,3]));
	@arg[4,5] = @arg[2,3] if (scalar(@arg)==2+$n{@arg[0]});
	${${$::EMCCharmm::Repair{@arg[1]}}{bond}}{$key} = join("\t", arrange(@arg[4,5]));
      }
    } elsif ($read==15) {					# patch
      my ($residue, $n) = split(":", shift(@arg));
      error_line($line, "no wildcards allowed\n") if ($residue =~ m/\*|\?/);
      my $patch = $::EMCCharmm::Residue{patch};
      my $ptr = $patch->{$residue} = {n => $n = ($n ? $n:1)};
      push(@{$::EMCCharmm::Residue{patches}}, $residue);
      $ptr->{delete} = [];
      while (@arg[-1] =~ m/^\+|^\-/) { 
	unshift(@{$ptr->{delete}}, pop(@arg));
      }
      $ptr->{core} = [splice(@arg,0,$n)];
      $ptr->{term} = [@arg];
    } elsif ($read==16) {					# wildcard
      $::EMCCharmm::ResidueWildcard = shift(@arg);
      while (scalar(@arg)>1) {
	my @atom = splice(@arg,0,2);
	$::EMCCharmm::Residue{wildcard}->{@atom[0]} = @atom[1];
      }
    } elsif ($read==17) {					# exclude
      foreach (@arg) { 
	my $residue = trim($_);
	my $ptr = $::EMCCharmm::Residue{exclude};
	push(@{$::EMCCharmm::Residue{excludes}}, $residue);
	$ptr->{$residue} = 1;
      }
      $::EMCCharmm::Flag{exclude} = 1;
    } elsif ($read==29) {					# include
      foreach (@arg) { 
	my $residue = trim($_);
	my $ptr = $::EMCCharmm::Residue{include};
	push(@{$::EMCCharmm::Residue{includes}}, $residue);
	$ptr->{$residue} = 1;
	$::EMCCharmm::Residue{rank}->{$residue} = $rank++;
      }
      $::EMCCharmm::Flag{include} = 1;
    } elsif ($read==18) {					# delete
      $::EMCCharmm::RuleDelete{join("\t", @arg[0,1])} = 1;
    } elsif ($read==19) {					# ring
      foreach (@arg) { 
	my @a = split(":");
	my $resi = shift(@a);
	my $type = shift(@a);
	my $key	= join("\t", $type, $resi);
	${$::EMCCharmm::Ring{$key}}{n} = 1;
	${$::EMCCharmm::Ring{$key}}{aromatic} = 0;
	foreach (@a) {
	  ${$::EMCCharmm::Ring{$key}}{aromatic} = 1 if ($_ eq "a");
	  ${$::EMCCharmm::Ring{$key}}{n} = eval($_) if ($_ ne "a");
	}
	${$::EMCCharmm::Ring{$key}}{n} = 1 if (${$::EMCCharmm::Ring{$key}}{n}<1);
      }
    } elsif ($read==22) {					# order
      my $key = join("\t", arrange(@arg[0,1]));
      $::EMCCharmm::BondOrder{$key} = @arg[2];
    } elsif ($read==23) {					# uncouple
      my $key = join("\t", arrange(@arg[0,1]));
      $::EMCCharmm::BondUncouple{$key} = 1;
    } elsif ($read==24) {					# increment
      my $std = join("\t", @arg[0,1]);
      my $key = join("\t", arrange(@arg[0,1]));
      $::EMCCharmm::IncrementsRead{$key} = 
	["-", join("\t", $std ne $key ? @arg[3,2] : @arg[2,3])];
    } elsif ($read==25) {					# retype
      my $key = join("\t", @arg[0,1]);
      $::EMCCharmm::Retype{$key} = @arg[2];
    } elsif ($read==27) {					# average
      if (!defined($::EMCCharmm::Element{@arg[1]})) {
	error_line($line, "undefined target type '@arg[1]'\n");
      }
      push(@::EMCCharmm::Average, [@arg]);
    } elsif ($read==28) {					# combine
      my $target = shift(@arg);
      if (!defined($::EMCCharmm::Element{$target})) {
	error_line($line, "undefined target type '$target'\n");
      }
      foreach (@arg) { 
	if (defined($::EMCCharmm::Combine{$_}) &&
	    $::EMCCharmm::Combine{$_} ne $target) {
	  error_line($line, "combination id '$_' already defined\n");
	}
	$::EMCCharmm::Combine{$_} = $target;
      }
    } elsif ($read==5) {					# extra
      $::EMCCharmm::Extras{$arg[0]} = join("\t", @arg);
    } elsif ($read==20) {					# nonbond
      my $key = join("\t", arrange(@arg[0,1]));
      $::EMCCharmm::Nonbond{$key} = join("\t", @arg[2,3,4,5]);
    } elsif ($read==6) {					# bond
      my $key = join("\t", arrange(splice(@arg,0,2)));
      my $k = join("\t", @arg);
      if ($key =~ m/\*/) {
	$::EMCCharmm::BondAuto{$key} = $k;
      } else {
	$::EMCCharmm::Bond{$key} = $k;
      }
    } elsif ($read==7) {					# angle
      my $key = join("\t", arrange(splice(@arg,0,3)));
      my $k = join("\t", @arg);
      if ($key =~ m/\*/) {
	$::EMCCharmm::AngleAuto{$key} = $k;
      } else {
	$::EMCCharmm::Angle{$key} = $k;
      }
    } elsif ($read==8) {					# torsion
      my $key = join("\t", arrange(splice(@arg,0,4)));
      my $k = join("\t", @arg);
      if ($key =~ m/\*/) {
	$::EMCCharmm::TorsionAuto{$key} = $k;
      } else {
	$::EMCCharmm::Torsion{$key} = $k;
      }
    } elsif ($read==9) {					# improper
      my $key = join("\t", arrange_imp(splice(@arg,0,4)));
      my $reverse = pop(@arg);
      my $k = join("\t", @arg);
      if ($key =~ m/\*/) {
	$::EMCCharmm::ImproperAuto{$key} = $k;
	$::EMCCharmm::ImproperReverseAuto{$key} = $reverse;
      } else {
	$::EMCCharmm::Improper{$key} = $k;
	$::EMCCharmm::ImproperReverse{$key} = $reverse;
      }
    }
  }
}


# handling of parameter file

sub get_header {
  my $line = 0;
  my $mode = 0;
  my $read = 0;
  my $flag = 1;
  my $version;
  my $text;

  undef @::EMCCharmm::References;

  foreach (keys(%::EMCCharmm::Journals)) {
    my $key = $_; $key =~ s/ //g; $::EMCCharmm::Journals{lc($key)} = $::EMCCharmm::Journals{$_};
  }
  seek($::EMCCharmm::PARMS, 0, 0);
  while (<$::EMCCharmm::PARMS>) {
    chop; ++$line;
    my @arg = split(" ");
    if (@arg[0] eq "") { ++$mode; next; }
    if ($mode == 0) {
      foreach (@arg) { $_ =~ s/,//g; }
      $flag = 0 if ($line == 1 && substr(@arg[1],0,3) eq "---");
      if ($flag) {
	if ($line == 1) {
	  $version = @arg[substr(@arg[0],1,1) eq ">" ? 1 : 2];
	} elsif ($line == 3) {
	  $version = "$version, @arg[1,2]";
	}
      } elsif ($version eq "") {
	if ($line == 2 && scalar(@arg)>2) {
	  $version = (split(":", @arg[1]))[0].", v@arg[-2]";
	}
      }
    } elsif ($mode == 1) {
      #error("comment expected in line $line of parameters\n") 
      last if (substr(@arg[0],0,1) ne "!");
      @arg[0] = substr(@arg[0], 1);
      if (@arg[0] eq "") {
	if ($read>1) {
	  $text = (split("\\. ", $text))[-1];
	  substr($text, -1, 1) = " " if (substr($text, -1, 1) eq ".");
	  my $m = ($text =~ m/\(/ ? 1 : 0);
	  $text =~ s/\(//g; $text =~ s /\)//g; $text =~ s/: /:/g;
	  my @arg = split(", ", $text);
	  if ($m) {
	    my @tmp = split(" ", $text);
	    @arg = (join(" ", @tmp[-6 .. -3]), @tmp[-2, -1]);
	  }
	  if (defined($::EMCCharmm::Journals{@arg[-4]})) {
	    @arg = (@arg[-4], @arg[-3], "@arg[-2]:@arg[-1]");
	  }
	  my $journal = @arg[-3];
	  my $m = (@arg[-2] =~ m/:/);
	  my $year = @arg[$m ? -1 : -2]; $year =~ s/ //g;
	  my $volume = @arg[$m ? -2 : -1];
	  my @tmp = split(":", $volume);
	  $volume = @tmp[0];
	  my $page = (split("-", @tmp[1]))[0];
	  $journal = lc($journal); $journal =~ s/ //g;
	  if (defined($::EMCCharmm::Journals{$journal})) {
	    push(@::EMCCharmm::References, join("\t",
		$year, $::EMCCharmm::Journals{$journal}, $volume, $page));
	  }
	}
	$text = "";
	$read = 0; next;
      }
      $text = join(" ", $read++ ? ($text, @arg) : @arg);
    } else { 
      last;
    }
  }
  if ($::EMCCharmm::FFVersion eq "UNKNOWN") {
    my @versions = ();
    foreach (@::EMCCharmm::TopologyIndex) {
      seek($::EMCCharmm::TOPOL, $_, SEEK_SET);
      while (<$::EMCCharmm::TOPOL>) {
	chop; ++$line;
	my @arg = split(" ");
	next if (substr(@arg[0],0,1) eq "*");
	push(@versions, "v@arg[0].@arg[1]");
	last;
      }
    }
    $version = join(", ", @versions);
  }
  $::EMCCharmm::FFVersion = $version if ($version ne "");
}


sub check {
  my $arg = shift(@_);
  foreach (@_) { return 1 if ($arg eq $_); }
  return 0;
}


sub get_parameters {
  my $target;
  my $n = 0;
  my $last = 0;
  my $mode = 0;
  my $cmap = 0;
  my $key = "";
  my %modes = (
    #"NONBONDED" => 1,
    "CUTNB" => 1, "BONDS" => 2, "ANGLES" => 3, "DIHEDRALS" => 4,
    "IMPROPERS" => 5, "IMPROPER" => 5, "CMAP" => 6, "NBFIX" => 7,
    "RESI" => 8, "PRES" => 8);
  my @types = (
    "none", "nonbonded", "bonds", "angles", "dihedrals", "impropers", "cmap", "nbfix");
  my @ntypes = (0, 1, 2, 3, 4, 4, 0, 0);

  seek($::EMCCharmm::PARMS, 0, 0);
  while (<$::EMCCharmm::PARMS>) {
    my @arg = split(" ", uc($_));
    if (@arg[0] eq "") { 
      next if (!$cmap);
      next if ($key eq "");
      $cmap = 0 if (scalar(@{$::EMCCharmm::CMAP{$key}{data}}) == $n*$n);
      next;
    }
    elsif (@arg[0] eq "END") { $mode = 0; next; }
    elsif (@arg[0] eq "HBOND") { next; }
    elsif (@arg[0] eq "NONBONDED") { next; }
    elsif ($mode == 8) { next; }
    elsif (substr(@arg[0],0,1) eq "!") { next; }

    if (defined($modes{$arg[0]})) {
      $mode = $modes{$arg[0]};
      info("reading $types[$mode]\n") if ($mode>0 && $mode<8);
      $key = "";
      next;
    }

    my @prm = ();
    foreach (@arg) { last if (substr($_,0,1) eq "!"); push(@prm, $_); }
    foreach (@prm[$ntypes[$mode] .. $#prm]) { $_ = sprintf("%.10g", $_); }
    if ($mode == 1) {						# cutnb
      push(@prm, @prm[1 .. 3]) if (scalar(@prm)==4);
      my $key = join("\t", arrange(@prm[0,0]));
      $::EMCCharmm::Nonbond{$key} = join("\t", list_round(@prm[3,2,6,5]));
    } elsif ($mode == 2) {					# BONDS
      my $key = join("\t", arrange(@prm[0 .. 1]));
      $::EMCCharmm::Bond{$key} = join("\t", list_round(@prm[2 .. $#prm]));
    } elsif ($mode == 3) {					# ANGLES
      my $key = join("\t", arrange(@prm[0 .. 2]));
      push(@prm, 0, 0) if (scalar(@prm)==5);
      $::EMCCharmm::Angle{$key} = join("\t", list_round(@prm[3 .. $#prm]));
    } elsif ($mode == 4) {					# DIHEDRALS
      foreach (@prm[0 .. 3]) { $_ = $::EMCCharmm::Wildcard if ($_ eq "X"); }
      my $key = join("\t", arrange(@prm[0 .. 3]));
      if ($key =~ m/\*/) {
	my $t = $::EMCCharmm::TorsionAuto{$key};
	$::EMCCharmm::TorsionAuto{$key} = (
	  $t eq "" ? "": "$t\t").join("\t", list_round(@prm[4 .. $#prm]));
	next;
      }
      my $t = $::EMCCharmm::Torsion{$key};
      $::EMCCharmm::Torsion{$key} = (
	$t eq "" ? "": "$t\t").join("\t", list_round(@prm[4 .. $#prm]));
    } elsif ($mode == 5) {					# IMPROPERS
      foreach (@prm[0 .. 3]) { $_ = $::EMCCharmm::Wildcard if ($_ eq "X"); }

      my $i; for ($i=0; $i<2; ++$i) {				# central can be
	my $key = join("\t", 					# first or last
	  arrange_imp($i ? reverse(@prm[0 .. 3]) : @prm[0 .. 3]));
	next if (!check(substr($key,0,1), ("C","N")));
	if ($key =~ m/\*/) {
	  $::EMCCharmm::ImproperAuto{$key} = join("\t", list_round(@prm[4,6]));
	  $::EMCCharmm::ImproperReverseAuto{$key} = 0;
	} else {
	  $::EMCCharmm::Improper{$key} = join("\t", list_round(@prm[4,6]));
	  $::EMCCharmm::ImproperReverse{$key} = 1;
	}
	last;
      }
    } elsif ($mode == 6) {					# CMAP
      if ($cmap) {
	next if ($key eq "");
	push(@{$::EMCCharmm::CMAP{$key}{data}}, @prm);
      } elsif (scalar(@arg) == 9) {
	$cmap = 1;
	$n = pop(@arg);
	$key = join("\t", @arg);
	$::EMCCharmm::CMAP{$key} = {n => $n, data => []};
      }
    }
  }
}


# topology: mass handling

@::EMCCharmm::ElementTable = (
     [1, "H", 1.0079], [1, "D", 2.0141], [2, "He", 4.0026],
     [3, "Li", 6.941], [4, "Be", 9.0122], [5, "B", 10.811],
     [6, "C", 12.011], [7, "N", 14.0067], [8, "O", 15.9994],
     [9, "F", 18.998], [10, "Ne", 20.18], [11, "Na", 22.99], 
     [12, "Mg", 24.305], [13, "Al", 26.982], [14, "Si", 28.085], 
     [15, "P", 30.974], [16, "S", 32.066], [17, "Cl", 35.4527], 
     [18, "Ar", 39.948], [19, "K", 39.098], [20, "Ca", 40.078], 
     [21, "Sc", 44.956], [22, "Ti", 47.867], [23, "V", 50.942], 
     [24, "Cr", 51.996], [25, "Mn", 54.938], [26, "Fe", 55.845], 
     [27, "Co", 58.933], [28, "Ni", 58.693], [29, "Cu", 63.546], 
     [30, "Zn", 65.39], [31, "Ga", 69.723], [32, "Ge", 72.61], 
     [33, "As", 74.922], [34, "Se", 78.96], [35, "Br", 79.904],
     [36, "Kr", 83.8], [37, "Rb", 85.468], [38, "Sr", 87.62],
     [39, "Y", 88.906], [40, "Zr", 91.224], [41, "Nb", 92.906],
     [42, "Mo", 95.94], [43, "Tc", 97.907], [44, "Ru", 101.07],
     [45, "Rh", 102.906], [46, "Pd", 106.42], [47, "Ag", 107.868],
     [48, "Cd", 112.411], [49, "In", 114.818], [50, "Sn", 118.71],
     [51, "Sb", 121.76], [52, "Te", 127.6], [53, "I", 126.904],
     [54, "Xe", 131.29], [55, "Cs", 132.905], [56, "Ba", 137.327],
     [57, "La", 138.905], [58, "Ce", 140.116], [59, "Pr", 140.908],
     [60, "Nd", 144.242], [61, "Pm", 144.913], [62, "Sm", 150.36],
     [63, "Eu", 151.964], [64, "Gd", 157.25], [65, "Tb", 158.925],
     [66, "Dy", 162.5], [67, "Ho", 164.93], [68, "Er", 167.259],
     [69, "Tm", 168.934], [70, "Yb", 173.04], [71, "Lu", 174.967],
     [72, "Hf", 178.49], [73, "Ta", 180.948], [74, "W", 183.84],
     [75, "Re", 186.207], [76, "Os", 190.23], [77, "Ir", 192.217],
     [78, "Pt", 195.084], [79, "Au", 196.967], [80, "Hg", 200.59],
     [81, "Tl", 204.383], [82, "Pb", 207.2], [83, "Bi", 208.98],
     [84, "Po", 208.982], [85, "At", 209.987], [86, "Rn", 222.018],
     [87, "Fr", 223.02], [88, "Ra", 226.0254], [89, "Ac", 227.027],
     [90, "Th", 232.038], [91, "Pa", 231.036], [92, "U", 238.029],
     [93, "Np", 237.048], [94, "Pu", 244.064], [95, "Am", 243.061],
     [96, "Cm", 247.07], [97, "Bk", 247.07], [98, "Cf", 251.08],
     [99, "Es", 252.083], [100, "Fm", 257.095], [101, "Md", 258.098],
     [102, "No", 259.101], [103, "Lr", 262.11], [104, "Rf", 263.113],
     [105, "Db", 262.114], [106, "Sg", 266.122], [107, "Bh", 264.1247],
     [108, "Hs", 269.134], [109, "Mt", 268.139], [110, "Ds", 272.146],
     [111, "Rg", 272.154], [112, "Cn", 277], [113, "Uut", 284],
     [114, "Uuq", 289], [115, "Uup", 288], [116, "Uuh", 292],
     [117, "Uus", 294], [118, "Uuo", 294]
   );

sub get_element {
  my $mass = shift(@_);

  foreach (@::EMCCharmm::ElementTable) {
    return @{$_}[1] if (abs(@{$_}[2]-$mass)<0.02);
  }
  return "?";
}


sub get_mass {
  my $element = shift(@_);

  foreach (@::EMCCharmm::ElementTable) {
    return @{$_}[2] if (@{$_}[1] eq $element);
  }
  return -1;
}


sub set_masses {
  my $order = shift(@_);
  my $index = shift(@_);
  my $type = shift(@_);
  my $mass = sprintf("%.10g", shift(@_));
  my $element = shift(@_); 
  shift(@_) if (@_[0] eq "!");
  my @a = (); foreach (@_) { last if (substr($_,0,1) eq "!"); push(@a, $_); }
  my $comment = ucfirst(join(" ", @a));

  if ($element eq "" || $element eq "!") {
    $element =
      defined($::EMCCharmm::Element{$type}) ? $::EMCCharmm::Element{$type} :
      get_element($mass);
  }
  $element = ucfirst(lc($element));
  $comment = "Element $element" if ($comment eq "");
  $::EMCCharmm::Masses{$type} = {
    data => [$mass, $element, $index, $comment], order => $order};
  $::EMCCharmm::Element{$type} = $element;
  $::EMCCharmm::Charge{$type} = 0.0;
}


sub cpy_masses {
  my $src = shift(@_);
  my $dest = shift(@_);
  my $mass = shift(@_);
  my $element = shift(@_); 
  my $nconns = shift(@_);
  my $index = shift(@_);
  shift(@_) if (@_[0] eq "!");
  my $comment = ucfirst(join(" ", @_));

  if (!defined($::EMCCharmm::Masses{$src})) {
    error("cpy_masses: type '$src' does not exist.\n");
  }

  my @orig = @{$::EMCCharmm::Masses{$src}->{data}};
  my $order = $::EMCCharmm::Masses{$src}->{order};

  $mass = @orig[0] if ($mass eq "" || $mass eq "-");
  $element = @orig[1] if ($element eq "" || $element eq "-");
  $index = @orig[2] if ($index eq "" || $index eq "-");
  $comment = @orig[3] if ($comment eq "" || $comment eq "-");
  $order = $::EMCCharmm::MassOrder if ($src ne $dest);
  $::EMCCharmm::NConnects{$dest} = $nconns;
  return ($dest, $mass, $element, 0, $index, $comment, $order);
}


sub new_type {
  my $type = shift(@_);
  my $occurence = shift(@_);
  my $neighborhood = shift(@_);
  my $ptr = $occurence->{$type};
  my $n = defined($ptr) ? $ptr->{n} : 0;
  my $typen;

  #if (defined($ptr->{neighborhood}->{$neighborhood})) {
  #  $n = $ptr->{neighborhood}->{$neighborhood};
  #}
  $type .= "_" if (length($type)==1);
  for (; $n<52; ++$n) {
    $typen = $type.chr($n>=26 ? $n+ord('a')-26 : $n+ord('A'));
    last if (!defined($::EMCCharmm::Masses{$typen}));
  }
  if ($n==52) {
    error("new type out of range (n>52)\n");
  }
  return $typen;
}


sub create_mass_entry {
  my $type = shift(@_);
  my $residue = shift(@_);
  my $atoms = shift(@_);
  my $typen = shift(@_);
  my $occurence = \%::EMCCharmm::Occurence;
  my $order = $::EMCCharmm::MassOrder;

  $typen = new_type($type, $occurence) if ($typen eq "");
  my @tmp = cpy_masses($type, $typen);
  $::EMCCharmm::TypeDefined{$type}->{$typen} = 1;
  $::EMCCharmm::TypeCreated{$typen} = 1;


  @tmp[5] .= " ($residue)" if ($residue ne "");

  set_masses($order, @tmp[4,0,1,2,5]);		# mass table entry

  foreach (\%::EMCCharmm::Equivalence,		# equivalence table entry
	   \%::EMCCharmm::EquivalenceAuto) {
    if (defined($_->{$type})) {
      @tmp = @{$_->{$type}};
    } else {
      @tmp = ($type, $type, $type, $type, $type, $type);
    }
    @tmp[1] = $typen;
    $_->{$typen} = [@tmp];
  }
  return $typen if ($residue eq "");
  
  foreach (@{$atoms}) {				# residue atom retypes
    my $key = join("\t", $residue, $_);
    $::EMCCharmm::Retype{$key} = $typen;
  }
  return $typen;
}


sub get_masses {
  my $line = 0;
  my $index = 0;
  my $fmass = 0;

  seek($::EMCCharmm::TOPOL, 0, 0);
  info("reading masses\n");
  while (<$::EMCCharmm::TOPOL>) {
    ++$line;
    my @arg = split(" ");
    if (uc(@arg[0]) eq "END") {
      ++$::EMCCharmm::MassOrder if ($fmass);
      $fmass = 0;
    }
    next if (uc(@arg[0]) ne "MASS");
    $fmass = 1;
    error_line($line, "missing mass\n") if (scalar(@arg)<2);
    if ($::EMCCharmm::TypeNChars<length(@arg[2])) {
      $::EMCCharmm::TypeNChars = length(@arg[2]);
    }
    shift(@arg);
    if (@arg[0]<0) {
      my $i = -1;
      if (defined($::EMCCharmm::Masses{@arg[1]})) {
	$i = @{$::EMCCharmm::Masses{@arg[1]}->{data}}[2]
      }
      @arg[0] = $i<0 ? ++$index : $i;
    }
    set_masses($::EMCCharmm::MassOrder, @arg);
  }
}


# topology: rule handling

sub create_residue_index {
  my ($type, $residue, %index);
  my $line = 0;
  my $rank = 0;

  info("creating residue index\n");
  seek($::EMCCharmm::TOPOL, 0, SEEK_SET);
  while (<$::EMCCharmm::TOPOL>) {
    my @arg = split(" ", uc($_));
    @arg[0] = substr(@arg[0],0,4);
    $residue = @arg[1];
    if (@arg[0] eq "RESI") {
      $type = defined($::EMCCharmm::Residue{exclude}->{$residue}) ? 0 : 1;
    } elsif (@arg[0] eq "PRES") {
      $type = defined($::EMCCharmm::Residue{exclude}->{$residue}) ? 0 : 
	      defined($::EMCCharmm::Residue{patch}->{$residue}) ? 2 : 0;
    } else {
      next;
    }
    #tprint(__LINE__.":", $residue, $type);
    if ($::EMCCharmm::Flag{include}) {
      $type = 0 if (!defined($::EMCCharmm::Residue{include}->{$residue}));
    }
    next if (!$type);
    push(@{$::EMCCharmm::Residue{residues}}, $residue);
    $::EMCCharmm::Residue{index}->{$residue} = tell($::EMCCharmm::TOPOL);
    $::EMCCharmm::Residue{charge}->{$residue} = $arg[2];
    $::EMCCharmm::Residue{type}->{$residue} = $type;
    if (!$::EMCCharmm::Flag{include}) {
      $::EMCCharmm::Residue{rank}->{$residue} = $rank++;
    }
    $type = 0;
  }
}


sub formal_charge {
  my $residue = shift(@_);
  my $atom = shift(@_);
  my $connect = shift(@_);
  my $bond = shift(@_);
  my $type = shift(@_);
  my $element = $::EMCCharmm::Element{$type->{$atom}};
  my $aromatic = aromatic($type->{$atom}, $residue);

  return $::EMCCharmm::Residue{charge}->{$residue} if (!%{$bond});
  $element =~ s/\[|\]//g;
  if ($element eq "N" || $element eq "O" ||
      $element eq "S" || $element eq "P") {
    my @connects = @{$connect->{$atom}};
    my $nconnects = scalar(@connects);
    my $charge = 0;

    foreach (@connects) {
      next if ($aromatic && $element eq "N" &&
	  $::EMCCharmm::Element{$type->{$_}} eq "H");
      $charge += $bond->{join("\t", arrange($atom, $_))};
    }
    $charge -= 5 if ($element eq "P");
    $charge -= 3 if ($element eq "N");
    $charge = 0 if ($element eq "N" && $aromatic && $charge == 1);
    $charge -= $aromatic ? 3 : 2 if ($element eq "O");
    if ($element eq "S") {
      --$charge if ($aromatic);
      if ($nconnects<=2) { $charge -= 2 }
      elsif ($nconnects==3) { $charge -= 4 }
      else { $charge -= 6; }
    }
    return $charge;
  }
  return 0;
}


sub element {
  my $residue = shift(@_);
  my $atom = shift(@_);
  my $connect = shift(@_);
  my $bond = shift(@_);
  my $type = shift(@_);
  my $charge = shift(@_);
  if (!defined($connect->{$atom})) {
    error("undefined atom '$atom' in residue '$residue'.\n");
  }
  my @connects = @{$connect->{$atom}};
  my $nconnects = scalar(@connects);
  my $element = $::EMCCharmm::Element{$type->{$atom}};
  my $aromatic = aromatic($type->{$atom}, $residue);

  if ($element eq "N" || $element eq "O" ||
      $element eq "S" || $element eq "P") {
    $charge = 0;
    foreach (@connects) {
      $charge += $bond->{join("\t", arrange($atom, $_))};
    }
    $charge -= 5 if ($element eq "P");
    $charge -= 3 if ($element eq "N");
    $charge = 0 if ($element eq "N" && $aromatic && $charge == 1);
    $charge -= $aromatic ? 3 : 2 if ($element eq "O");
    if ($element eq "S") {
      --$charge if ($aromatic);
      if ($nconnects<=2) { $charge -= 2 }
      elsif ($nconnects==3) { $charge -= 4 }
      else { $charge -= 6; }
    }
  }
  $charge = 0 if (int($charge) != $charge);
  $element .= abs($charge)>1 ? sprintf("%s%g", $charge>0 ? "+" : "", $charge) :
	      $charge<0 ? "-" : $charge>0 ? "+" : "";
  $element = "[$element]" if (length($element)>1);
  $element = lc($element) if ($aromatic);
  return $element ne "" ? $element : "*";
}


%::EMCCharmm::BondType = ("-" => 1, ":" => 1.5, "=" => 2, "#" => 3);
%::EMCCharmm::BondChar = (1 => "", 1.5 => ":", 2 => "=", 3 => "#");

sub set_bond {
  my $bond = shift(@_);
  my @a = arrange(splice(@_,0,2));

  return $::EMCCharmm::BondChar{$bond->{join("\t", @a)}};
}


sub ring_type {
  my @c = split(@_[0]);
  return 0 if (@c[0] ne "^");
  return @c[1] if (@c[1] ge "3" && @c[1] le "9");
  return 1;
}


%::EMCCharmm::SortAtom = ();

sub atom_compare {
  #my $x = lc($::EMCCharmm::SortAtom{@_[0]});
  #my $y = lc($::EMCCharmm::SortAtom{@_[1]});
  my $x = $::EMCCharmm::SortAtom{@_[0]};
  my $y = $::EMCCharmm::SortAtom{@_[1]};
  my $ax = $x;
  my $ay = $y;
  $x =~ s/\[|\]//g;
  $y =~ s/\[|\]//g;
  return 1 if ($x eq "*");
  return -1 if ($y eq "*");
  return 1 if ($x eq "h");
  return -1 if ($y eq "h");
  my $bx = $::EMCCharmm::BondType{substr($x,0,1)}; $x = substr($x,1) if ($bx);
  my $by = $::EMCCharmm::BondType{substr($y,0,1)}; $y = substr($y,1) if ($by);
  my $cx = ring_type(substr($x,0,2)); $x = substr($x,$cx>1 ? 2 : 1) if ($cx);
  my $cy = ring_type(substr($y,0,2)); $y = substr($y,$cy>1 ? 2 : 1) if ($cy);
  return $x lt $y ? -1 : 1 if ($x ne $y);
  return $cx>$cy ? -1 : 1 if ($cx != $cy);
  return $bx>$by ? -1 : 1 if ($bx != $by);
  return $ax lt $ay ? -1 : 1 if ($ax ne $ay);
  return 0;
}


sub sort_compare {
  return atom_compare($a, $b);
}


sub create_rule {
  my $residue = shift(@_);
  my $atom = shift(@_);
  my $igroup = shift(@_);
  my $rule = shift(@_);
  my $group = shift(@_);
  my $type = shift(@_);
  my $connect = shift(@_);
  my $bond = shift(@_);
  my $charge = shift(@_);
  my $visited = shift(@_);
  my $rules = shift(@_);
  my $depth = shift(@_);
  my $level = shift(@_);
  my $last = shift(@_);
  my $flag = $atom eq "N";
  my %sorted = ();
  my %rules_loc;

  $visited->{$atom} = 0;		# 0: allow for redundancy in rings
  if ($depth && $rule ne "*") {		# EMC <= 9.3.9 does as well
    my $arom = aromatic($type->{$atom}, $residue);
    my $b = $::EMCCharmm::BondChar{$::EMCCharmm::BondType{substr($rule,0,1)}};
    my $a = aromatic($type->{$atom}, $residue);
    my $r = $a ? 0 : ring($type->{$atom}, $residue);
    
    $b .= $r==1 ? "^" : "^$r" if ($r);
    $rule = $b.element(
		  $residue, $atom, $connect, $bond, $type, $charge->{$atom});
    if (1 || $igroup<0 || $igroup == $group->{$atom}) {
      my $double = 0;

      foreach (@{$connect->{$atom}}) {
	my $c = $_; $c =~ s/[+|-]//g;
	my $a = aromatic($type->{$c}, $residue);
	my $r = $a ? 0 : ring($type->{$c}, $residue);
	   $r = $r ? $r==1 ? "^" : "^$r" : "";
	my $t = lc($::EMCCharmm::Element{$type->{$c}}); 
	   $t = ucfirst($t) if ($a);
	my $e = element($residue, $c, $connect, $bond, $type, $charge->{$c});
	my $b = set_bond($bond, $atom, $c).$r;
	$t = "*" if ($t eq "");
	$b = ":" if ($arom && ($e eq "c" || $e eq "n"));
	$sorted{$c} = $b.($e =~ m/\[/ ? "[$t]" : $t);
	$rules_loc{$c} = $b.$e;
	$double = 1 if ($b eq "=");
      }
      %::EMCCharmm::SortAtom = %sorted;
      my (@subs, $ls);
      my @sort = sort(sort_compare keys(%sorted));
      foreach (@sort) {
	my @arg = split("\t");
	next if ($visited->{@arg[0]});
	next if (@arg[0] eq $last);
	if (@arg[0] eq "+N" || @arg[0] eq "-C") {
	  push(@subs, "(".substr(@arg[0],1).")");
	} else {
	  push(@subs, "(".
	    create_rule(
	      $residue, @arg[0], $igroup, $rules_loc{@arg[0]}, 
	      $group, $type, $connect, $bond, $charge, 
	      $visited, $rules, $depth-1, $level+1, $atom, $b).")");
	}
	$ls = $_;
      }
      foreach(sort(@subs)) { $rule .= $_ }
    }
  }
  return $rule;
}


sub sort_connect {
  return
    substr($a,0,1) ne "H" ?
      substr($b,0,1) ne "H" ?
	$a lt $b ? -1 : $a gt $b ? 1 : 0 :
      -1 :
    substr($b,0,1) eq "H" ? 
      $a lt $b ? -1 : $a gt $b ? 1 : 0 : 1;
}


sub order_connect {
  my $connect = shift(@_);

  foreach (keys(%$connect)) {
    $connect->{$_} = [sort(sort_connect @{$connect->{$_}})];
  }
}


sub smiles {
  my $residue = shift(@_);
  my $atom = shift(@_);
  my $connect = shift(@_);
  my $bond = shift(@_);
  my $type = shift(@_);
  my $charge = shift(@_);
  my $visited = shift(@_);
  my $link = shift(@_);
  my $h = shift(@_);
  my $element = element(
    $residue, $atom, $connect, $bond, $type, $charge->{$atom});
  my $smiles = $element;
  my @connects = @{$connect->{$atom}};
  my $nconnects = scalar(@connects);
  my $last = "";

  $visited->{$atom} = 1;
  foreach (@{$link->{$atom}}) { $smiles .=  $_>9 ? "%$_" : $_; }
  foreach (@connects) {
    my $b = set_bond($bond, $atom, $_);
    $b = "" if (aromatic($type->{$atom}, $residue) && 
		aromatic($type->{$_}, $residue));
    if ($visited->{$_}) {
      next;
    } elsif (substr($_,0,1) eq "-") {
      $smiles = "*".$smiles;
    } elsif (substr($_,0,1) eq "+") {
      $smiles .= "(*)";
    } elsif (!$h && 
	      $::EMCCharmm::Element{$type->{$_}} eq "H" && $element ne "n") {
      next;
    } elsif ($last eq "") {
      $last = $b.smiles($residue,
	$_, $connect, $bond, $type, $charge, $visited, $link, $h);
    } else {
      $smiles .= "(".$b.smiles($residue,
	$_, $connect, $bond, $type, $charge, $visited, $link, $h).")";
    }
  }
  return $smiles.$last;
}


sub set_links {
  my $atom = shift(@_);
  my $connect = shift(@_);
  my $visited = shift(@_);
  my $link = shift(@_);
  my $nlink = shift(@_);
  my $level = shift(@_);

  $link->{$atom} = [];
  $visited->{$atom} = $level;
  foreach (@{$connect->{$atom}}) {
    if ($visited->{$_}) {
      next if ($level-$visited->{$_}<2);
      push(@{$link->{$atom}}, $$nlink);
      push(@{$link->{$_}}, $$nlink++);
      next;
    } else {
      set_links($_, $connect, $visited, $link, $nlink, $level+1);
    }
  }
}


sub add_bonds {
  my $bond = shift(@_);
  my $type = shift(@_);
  my $wild = shift(@_);
  
  while (scalar(@_)) {
    my @arg = splice(@_, 0, 2);
    my @tmp = @arg;
    if ($wild) {
      foreach (@arg) {
	next if (!defined($::EMCCharmm::Residue{wildcard}->{$_}));
	$_ = $::EMCCharmm::Residue{wildcard}->{$_};
      }
    }
    my $key = join("\t", arrange(@arg));
    $bond->{$key} = $type if (!defined($bond->{$key}));
  }
}


sub delete_bonds {
  my $bond = shift(@_);
  my $id = shift(@_);
  
  foreach (keys(%{$bond})) {
    my @arg = split("\t", $_);
    delete($bond->{$_}) if ((@arg[0] eq $id) || (@arg[1] eq $id));
  }
}


sub get_residue {
  my $stop = 0;
  my $dpth = 0;

  my $marker = shift(@_);
  if (defined($::EMCCharmm::FFTrial)) {
    $stop = 1; $marker = $::EMCCharmm::FFTrial; $dpth = 2; 
  }

  my @arg = split("\\.", $marker);
  my $index =  join(".", @arg);
  my $residue = @arg[0];

  return undef if (!defined($::EMCCharmm::Residue{index}->{$residue}));
  if (defined($::EMCCharmm::FFTrial)) {
    debug("marker = $marker\n");
  }

  my $combine = scalar(@arg)>1 ? 1 : 0;
  my $depth = shift(@_); $depth = $dpth if ($dpth);
  my $fgroup = shift(@_);
  my $rules = shift(@_);
  my $selection = shift(@_);
  my $retype = shift(@_); 
     $retype = \%::EMCCharmm::Retype if (!defined($retype));
  my $set = scalar(@{$selection})>0 ? 0 : 1;
  my $patch = defined($::EMCCharmm::Residue{patch}->{$marker}) ?
	      $::EMCCharmm::Residue{patch}->{$marker} : 0;
  my $ncore = 1;
  my $nterm = 0;
  my %extra = (
    "C" => join("\t", arrange("C", "*")),
    "N" => join("\t", arrange("N", "*")));
  my %modes = (
    "RESI" => 1, "PRES" => 2, "GROU" => 3, "ATOM" => 4,
    "BOND" => 5, "DOUB" => 6, "IMPR" => 7, "CMAP" => 8,
    "DONO" => 9, "ACCE" => 10, "IC" => 11, "DELE" => 12,
    "END" => 13, "BILD" => 14);
  my %type = (
    "*" => "*");
  my %ntypes = ();
  my $igroup = -1;
  my %charge = ();
  my %formal = ();
  my %bond = ();
  my %improper = ();
  my %connect = ();
  my %nconnects = ();
  my %group = ();
  my $first = "";
  my @atom = ();
  my %atom = ();
  my %exclude = ();
  my %uncouple = ();
  my @list = ($residue);
  my $ilist = -1;
  my $nrepeat = 0;
  my $wild = 0;
  my $count = 0;
  my $delete = "";
  my @delete = ();
  my %link = ();
  my $offset = 0;
  my $mode;

  # prepare patch

  if ($patch) {						# check for patch
    $ncore = scalar(@{$patch->{core}});
    $nterm = scalar(@{$patch->{term}});
    @delete = @{$patch->{delete}};
    foreach (@delete) { $_ = "-" if ($_ eq ""); }
    $delete = @delete[0];
    unshift(@list, @{$patch->{core}}, @{$patch->{term}});
    $count = -1 if (scalar(@{$patch->{core}})<2 && $delete ne "-");
    $count = -1 if (scalar(@list)<3);
    $offset = $ncore+$count;
  }

  # gather residue information

  debug("residue:\n");
  tdebug("","$marker", @{$selection});
  foreach (@list) {					# loop over list
    my $resi = $_;
    my $self = $resi eq $residue;
    $nrepeat = 0 if ($self);
    my $core = (++$ilist>=$ncore)&&($ilist<$ncore+$nterm) ? 0 : 1;
    my $n = $core ? 1 : $nrepeat;
    my $i;

    for ($i=0; $i<$n; ++$i) {
      my %resi_bond;
      my $pre = ++$count<1 ? "" : $self ? "" : $count;
     
      #tprint(__LINE__.":", $resi, $pre) if ($::EMCCharmm::Flag{debug});
      seek($::EMCCharmm::TOPOL, $::EMCCharmm::Residue{index}->{$resi}, SEEK_SET);
      while (<$::EMCCharmm::TOPOL>) {			# interpret topology
	my @arg = (); 
	foreach (split(" ", uc($_))) { 
	  last if (substr($_,0,1) eq "!");
	  push(@arg, $_);
	}
	next if (!($mode = $modes{substr(shift(@arg),0,4)}));
	if ($mode==3) {					# GROUP
	  ++$igroup;
	} elsif ($mode==4) {				# ATOM
	  my $key = join("\t", $resi, @arg[0]);
	  next if ($::EMCCharmm::RuleDelete{$key});
	  if (defined($::EMCCharmm::Charge{$key})) {
	    @arg[2] = $::EMCCharmm::Charge{$key};
	  }
	  my $id = $pre.shift(@arg);
	  push(@atom, $id) if (!$atom{$id}); $atom{$id} = 1;
	  $first = $id if ($first eq "");
	  push(@{$selection}, $id) if ($set && $self);
	  if (!$patch) {
	    $bond{$extra{"C"}} = 1 if (join(" ", @arg[0,1]) eq "C C");
	    $bond{$extra{"N"}} = 1 if (join(" ", @arg[0,1]) eq "N NH1");
	  }
	  if ($wild && defined($::EMCCharmm::Residue{wildcard}->{$id})) {
	    $type{$id} = $::EMCCharmm::Residue{wildcard}->{$id}; shift(@arg);
	    $charge{$id} = "0"; shift(@arg);
	  } else {
	    my $key = "$residue\t$id";
	    $type{$id} = shift(@arg);
	    $charge{$id} = sprintf("%.10g", shift(@arg));
	  }
	  $group{$id} = $igroup<0 ? ++$igroup : $igroup;
	} elsif ($mode==5 || $mode==6 || $mode==8) {	# BOND, DOUBLE or CMAP
	  add_bonds(\%resi_bond, $mode==6 ? 2 : 1, $wild, @arg);
	} elsif ($mode==7) {				# IMPR
	  my @impropers = @arg;
	  while (scalar(@impropers)) {
	    my @arg = (splice(@impropers,0,4))[0,1,0,2,0,3];
	    add_bonds(\%resi_bond, 1, $wild, @arg);
	    $improper{join("\t", arrange(@arg[0,1,3,5]))} = 1;
	  }
	} elsif ($mode==9 || $mode==10 ) {		# DONO or ACCE
	  next;
	} elsif ($mode==11 || $mode==14 ) {		# IC or BILD
	    my @dihedrals = @arg[0,1,2,3];
	    while (scalar(@dihedrals)) {		# dihedral assumed
	      my @arg = splice(@dihedrals,0,4);
	      if (substr(@arg[2],0,1) eq "*") {		# improper if '*'-type
		@arg[2] = substr(@arg[2],1); @arg = @arg[2,0,2,1,2,3];
	      }
	      my $count = 0;
	      foreach (@arg) {
		my $c = substr($_ = uc($_),0,1);	# filter out issues
		if ($c eq "*") { $count = 2; last; }
		if ($c eq "+" || $c eq "-") { ++$count; next; }
		if (!defined($type{$_})) { $count = 2; last; }
	      }
	      next if ($count!=1);
	      add_bonds(\%resi_bond, 1, $wild, @arg);
	    }
	} elsif ($mode==12) {				# DELETE
	  if (shift(@arg) eq "ATOM") {
	    my $id = shift(@arg);
	    $exclude{$id} = 1;
	    delete_bonds(\%bond, $id);
	    delete($type{$id});
	    delete($group{$id});
	    delete($charge{$id});
	  }
	} else {					# RESI, PRES, etc.
	  my %ignore;					# at end of RESI etc.
	  my @bonds = ();				# read exit

	  foreach (@delete) {				# determine ignored	
	    my @a = split(":");
	    $ignore{@a[0]} = 1 if ((scalar(@a)<2 && $pre==1) || @a[1]==$pre);
	  }
	  foreach (sort(keys(%resi_bond))) {
	    my $atom;
	    my @arg = ();
	    my $flag = 0;
	    my @atoms = split("\t");
	    
	    foreach (@atoms) {
	      $flag = 1 if (defined($ignore{$_}));
	      $atom = $pre.$_ if (!($_ =~ m/^\+|^\-/));  
	    }
	    next if ($flag);
	    foreach (@atoms) {
	      my $id = $pre.$_;
	      if ($_ =~ m/^\+|^\-/) {
		if (defined($link{$atom})) {
		  $id = $link{$atom};
		} else {
		  $id = ($offset+$nrepeat+1).substr($_,1);
		  if ($self) {
		    if (!defined($link{$id})) { $id = $_; }
		  } else {
		    ++$nrepeat;
		    #tprint(__LINE__.":".$self, "link", $id, "<=>", $atom) if ($::EMCCharmm::Flag{debug});
		    $link{$id} = $atom;
		  }
		  $type{$id} = $type{$pre.substr($_,1)};
		}
	      }
	      push(@arg, $id);
	    }
	    #tprint(__LINE__.":", $_, "->", @arg) if ($::EMCCharmm::Flag{debug});
	    $bond{join("\t", arrange(@arg))} = $resi_bond{$_};
	  }
	  #debug("\n");
	  last;
	}
      }
    }
  }

  # post processing

  foreach (@::EMCCharmm::Average) {			# averaging
    my @arg = @{$_};
    my $resi = shift(@arg);
    my $target = shift(@arg);
    my $c = 0;
    my $n = 0;

    next if ($residue ne $resi);
    foreach (@arg) {
      my $id = $_;
      error("undefined averaging id '$id'\n") if (!defined($charge{$id}));
      $c += $charge{$id};
      ++$n;
    }
    $c /= $n if ($n);
    foreach (@arg) {
      my $id = $_;
      $type{$id} = $target;
      $charge{$id} = $c;
    }
  }
  
  foreach (sort(keys(%type))) {				# retypes and count
    my $t = $type{$_};
    my $key = join("\t", $residue, $_);
    ++$ntypes{$t};
    $::EMCCharmm::NConnects{$t} = 0 if (!defined($::EMCCharmm::NConnects{$t}));
    next if (!defined($retype->{$key}));
    $type{$_} = $retype->{$key};
  }
  
  if ($patch) {						# delete through connect
    if ($delete ne "" && $delete ne "-") {
      $exclude{$delete} = 1;
      delete_bonds(\%bond, $delete);
      delete($type{$delete});
      delete($group{$delete});
      delete($charge{$delete});
    }
  }
  
  foreach (keys(%bond)) {				# create connectivity
    my @arg = split("\t");
    push(@{$connect{@arg[0]}}, @arg[1]);
    push(@{$connect{@arg[1]}}, @arg[0]);
  }

  foreach (keys(%connect)) {
    my $t = $type{$_};
    my $c = $charge{$_};
    my $n = scalar(@{$connect{$_}});
    $nconnects{$_} = $n;
    $::EMCCharmm::NConnects{$t} = $n if ($n>$::EMCCharmm::NConnects{$t});
  }
  
  repair_bond($residue, \@atom, \%bond, \%type, \%connect, \%nconnects);

  if ((!defined($::EMCCharmm::Templates{$residue}) && $depth<2) || $stop) {
    my (%visited, %link);
    my $nlink = 1;					# create templates

    debug("\n");
    debug("connectivity:\n");
    order_connect(\%connect);
    foreach (@atom) {
      tdebug("", $_, @{$connect{$_}}) if (!$exclude{$_});
      $::EMCCharmm::Charge{$type{$_}} = $charge{$_} if (!scalar(@{$connect{$_}}));
    }
    if ($::EMCCharmm::Flag{debug}) {
      debug("\n"); debug("bonds: $depth\n"); 
      foreach (sort(keys(%bond))) {
	my @t = split("\t", $_); foreach (@t) { $_ = $type{$_}; }
	tdebug("", $_, @t, $bond{$_});
      }
    }
    if (!$patch) {
      set_links($first, \%connect, \%visited, \%link, \$nlink, 1);
      %visited = ();
      $::EMCCharmm::Templates{$residue} = smiles($residue,
	  $first, \%connect, \%bond, \%type, \%charge,
	  \%visited, \%link, 0, $residue eq "TIP3") if (!$patch);
      debug("\n");
      debug("template:\n");
      tdebug("", $::EMCCharmm::Templates{$residue});
    }
  }
  return (
    marker => $marker, residue => $residue, 
    fgroup => $fgroup, depth => $depth, stop => $stop, index => $index,
    selection => $selection, rules => $rules,
    charge => {%charge}, connect => {%connect},
    bond => {%bond}, type => {%type}, ntypes => {%ntypes}, group => {%group},
    retype => $retype
  );
}


# increment handling

sub assign_increments {
  my $r = shift(@_);
  my $pass = shift(@_);
  my $reverse = shift(@_);
  my $marker = $r->{marker};
  my @arg = split("\\.", $marker);
  my $residue = @arg[0];
  my $selection = $r->{selection};
  my $neighbor = $r->{neighbor}->{$residue};
  my $connect = $r->{connect};
  my $charge = $r->{charge};
  my $bond = $r->{bond};
  my $type = $r->{type};
  my $group = $r->{group};
  my $show = $::EMCCharmm::Flag{show};
  my $debug = $::EMCCharmm::Flag{debug};
  my $n = scalar(@{$selection});
  my @sel = sort(@{$selection});
  my $res_charge = $::EMCCharmm::Residue{charge}->{$residue};
  my $i = 0;
  my $ncharged = 0;
  my %increments;
  my %uncouple;
  my %formal;
  my %index;
  my @null;
  my %bonds;
  my %connected;
  my @m;

  debug("charge = $::EMCCharmm::Residue{charge}->{$residue}\n"); debug("\n");
  debug("charges:\n"); 
  #foreach (@sel = sort(keys(%{$charge}))) {
  foreach (sort(keys(%{$charge}))) {
    debug(join("\t", "", $_, $type->{$_}, $charge->{$_}), "\n");
  }
  debug("\n"); debug("formal charges:\n"); 
  foreach (sort(keys(%formal))) {
    debug(join("\t", "", $_, $type->{$_}, $group->{$_}, $formal{$_}), "\n");
  }
  return if (!%{$bond});
  foreach (@sel) {
    $formal{$_} = formal_charge($residue, $_, $connect, $bond, $type);
    ++$ncharged if ($formal{$_});
  }
  debug("\n"); debug("bonds:\n"); 
  foreach (sort(keys(%{$bond}))) {
    my @t = split("\t", $_); foreach (@t) { $_ = $type->{$_}; }
    tdebug("", $_, @t, $bond->{$_});
    my $key = join("\t", arrange(@t));
    my $hash = hash_entry(\%::EMCCharmm::BondUncouple, arrange(@t));
    $uncouple{$key} = 1 if (defined($hash));
  }
  if ($ncharged) {					# uncouple bonds
    my %allowed = ("c" => 1, "n" => 1, "p" => 1);

    foreach (sort(keys(%formal))) {
      next if (!$formal{$_});
      my $flag = 0;
      my $s = $_;
      foreach (@{$connect->{$s}}) {
	next if (!$formal{$_});
	my $key = join("\t", arrange($type->{$s}, $type->{$_}));
	$uncouple{$key} = $flag = 1;
	last;
      }
      if (!$flag) {
	foreach (@{$connect->{$s}}) {
	  next if (!defined($allowed{lc($::EMCCharmm::Element{$type->{$_}})}));
	  my $key = join("\t", arrange($type->{$s}, $type->{$_}));
	  $uncouple{$key} = 1;
	  last;
	}
      }
    }
  }
  if (%uncouple) {
    debug("\n"); debug("uncouple:\n");
    foreach (sort(keys(%uncouple))) { debug("\t$_\n"); }
  }
  debug("\n");
  foreach (@sel) {					# create bond list
    my $s = $_; foreach (@{$connect->{$s}}) {
      my $key = join("\t", arrange($s, $_));
      $connected{$_} = 1;
      $bonds{$key} = 1;
    }
  }
  foreach (sort(keys(%bonds))) {			# create index
    my $count = 0;
    my @arg = split("\t", $_);
    foreach (@arg) {
      last if (!defined($type->{$_}));
      last if ($type->{$_} eq "*");
      ++$count; $_ = $type->{$_};
    }
    next if ($count!=2);
    $index{join("\t", @arg)} = 1;
    $index{join("\t", reverse(@arg))} = 1;
  }
  my @keys = sort(keys(%index));
  foreach (@keys) { 
    $index{$_} = $i++; push(@null, 0);
  }
  my @total = @null;
  print("\n") if ($show);
  foreach (@sel) {					# create matrix
    my $s = $_; 
    next if (!defined($type->{$s}) || $type->{$s} eq "*");
    my @row = @null;
    foreach (@{$connect->{$_}}) {
      my $c = $_;
      next if (!defined($type->{$c}) || $type->{$c} eq "*");
      my $id = join("\t", $type->{$s}, $type->{$c});
      ++@total[$index{$id}];
    }
    foreach (sort(keys(%connected))) {
      my $c = $_; 
      next if (!defined($type->{$c}) || $type->{$c} eq "*");
      #next if ($type->{$s} eq $type->{$_});
      my $key = join("\t", arrange($s, $c));
      my $id = join("\t", $type->{$s}, $type->{$c});
      ++@row[$index{$id}] if (defined($bonds{$key}));
    }
    push(@row, eval(${$charge}{$s}));
    print(join("\t", __LINE__.":", @row), "\n") if ($show);
    push(@m, [@row]);
  }
  push(@total, eval($::EMCCharmm::Residue{charge}->{$residue}));
  BOND: foreach (sort(keys(%bonds))) {			# add extras
    my @arg = split("\t");
    foreach (@arg) { $_ =~ s/[+|-]//g; }
    my @grp = ($group->{@arg[0]}, $group->{@arg[1]});
    foreach (@arg) {					# convert to type
      next BOND if (!defined($type->{$_}));
      next BOND if (($_ = $type->{$_}) eq "*");
    }
    my $key = join("\t", arrange(@arg));
    my @id = ($index{join("\t", @arg)}, $index{join("\t", reverse(@arg))});
    my @row = (@null, 0);
    if (!defined($uncouple{$key})) {			# for charged asymmetry
      @row[@id[0]] = @row[@id[1]] = 1; push(@m, [@row]);# val + compl = 0
    }
    #my $inc = hash_entry(\%::EMCCharmm::IncrementsRead, arrange(@arg));
    #next if (!defined($inc));
    next if (!defined($::EMCCharmm::IncrementsRead{$key}));
    my $inc = $::EMCCharmm::IncrementsRead{$key};
    @arg = split("\t", $key);				# add already existing
    next if (@arg[0] eq "*" || @arg[1] eq "*");
    @id = ($index{join("\t", @arg)}, $index{join("\t", reverse(@arg))});
    #if (!defined($::EMCCharmm::Increments{$key})) {
    #  $::EMCCharmm::Increments{$key} = [$residue, $inc->[1]];
    #}
    my @v = split("\t", $inc->[1]);
    @row[@id[0]] = 1; @row[@id[1]] = 0; @row[-1] = @v[0]; push(@m, [@row]);
    tprint(__LINE__.":", @row) if ($show);
    @row[@id[0]] = 0; @row[@id[1]] = 1; @row[-1] = @v[1]; push(@m, [@row]);
    tprint(__LINE__.":", @row) if ($show);
  }
  my %increments;					# solve linear eqns

  print("\n") if ($show);
  debug("charge = $res_charge\n"); debug("\n");
  print("\n") if ($show);
  tprint(__LINE__.":", "MATRIX", $residue."\n") if ($show); 
  EMC::Matrix::mheader(\@keys) if ($show);
  EMC::Matrix::mreduce(\@m); push(@m, [@total]);
  EMC::Matrix::mprint(\@m) if ($show);
  
  my $flag = 0;
  my $sum = 0.0;
  my %issues = ();
  my $svd = 0;
  my $m;
  my $sol = EMC::Matrix::msolve(
    \@m, $::EMCCharmm::Round, $show, \%issues, \$svd, \$m);

  if ($svd)						# try rounding
  {
    my $i;
    my $round = $::EMCCharmm::SVDRound;
    my $nconnects = \%::EMCCharmm::NConnects;
    my %v;

    for ($i=0; $i<scalar(@{$sol}); ++$i) {
      $v{@keys[$i]} = round($sol->[$i], 10.0*$round);
    }
    for ($i=0; $i<scalar(@{$sol}); ++$i) {
      my @r = (round($sol->[$i], $round), round($sol->[$i], 0.1*$round));
      next if (@r[0] != @r[1]);
      
      my @arg = split("\t", @keys[$i]);
      next if ($nconnects->{@arg[0]}==1 || $nconnects->{@arg[1]}==1);

      my @id = ($index{join("\t", @arg)}, $index{join("\t", reverse(@arg))});
      my @row = (@null, 0);
      my @v = (round($v{join("\t", @arg)}), round($v{join("\t", reverse(@arg))}));

      @row[@id[0]] = 1; @row[@id[1]] = 0; @row[-1] = @v[0]; push(@m, [@row]);
      @row[@id[0]] = 0; @row[@id[1]] = 1; @row[-1] = @v[1]; push(@m, [@row]);
      round($v{join("\t", @arg)});

      my $s = EMC::Matrix::msolve(
	\@m, $::EMCCharmm::Round, $show, \%issues, \$svd, \$m);
      $sol = $s if (defined($s));
      last;
    }
  }

  if (!defined($sol)) {
    $sol = [];
    $flag = -1;
    for (my $i=0; $i<scalar(@{$m->[0]})-1; ++$i) { 
      push(@{$sol}, $m->[$i]->[-1]);
    }
  }
  if (defined($sol)) {					# transcribe solution
    for (my $i=0; $i<scalar(@{$sol}); ++$i) {
      $sum += $sol->[$i]*@total[$i];
    }							# allow for extra 10%
    $flag = 1 if (!$flag && abs($sum-@total[-1])<1.1*$::EMCCharmm::Round);
    $flag = 0 if ($flag<0);
    debug("sum:\n");
    tdebug("", $residue, @total[-1], $sum, abs(@total[-1]-$sum));
    debug("\n");
    for (my $i=0; $i<scalar(@{$sol}); ++$i) {
      my @arg = split("\t", @keys[$i]);
      my $key = join("\t", arrange(@arg));
      my $inc = $increments{$key};
      $increments{$key} = ($inc eq "" ? "" : "$inc\t").$sol->[$i];
      $increments{$key} .= "\t$sol->[$i]" if (@arg[0] eq @arg[1]);
    }
    debug("increments:\n");
    foreach (sort(keys(%increments))) {
      debug("\t$_\t$increments{$_}\n");
    }
    debug("\n");
    create_increment_tree(
      $residue, \%increments, \%bonds, $type, $reverse, 0) if ($flag);
  } 
  return $flag if ($flag || $pass);
  
  foreach (keys(%issues)) {
    $issues{join("\t", arrange(split("\t", @keys[$_])))} = 1;
    delete($issues{$_});
  }
  
  create_type_additions(
    $r, \%bonds, $type, \%formal, \%uncouple, \%issues);
  return $flag;
}


sub create_increment_tree {
  my $residue = shift(@_);
  my $increments = shift(@_);
  my $bonds = shift(@_);
  my $type = shift(@_);
  my $reverse = shift(@_);
  my $warning = shift(@_);

  foreach (sort(keys(%{$increments}))) {
    my $inc = $increments->{$_};
    my @list = list_atoms($_, $type, $bonds);
    my %data = (residue => $residue, atoms => {@list});
    my @arg = split("\t");
    my %change;
    if (defined($reverse)) {
      foreach (@arg) { 
	next if (!defined($reverse->{$_})); 
	$change{$_} = 1; $_ = $reverse->{$_};
      }
    }
    my $ptr = $::EMCCharmm::IncrementTree{@arg[0]}->{@arg[1]};

    $data{change} = {%change} if (%change);
    if (defined($ptr)) {
      if ($warning && !defined($ptr->{$inc})) {
	my $set = join(", ", @arg, split("\t", $inc));
	warning("residue $residue increment {$set} exists\n");
      }
      push(@{$ptr->{$inc}}, {%data});
    } else {
      $::EMCCharmm::IncrementTree{@arg[0]}->{@arg[1]}->{$inc} = [{%data}];
    }
  }
}


sub list_bonds {
  my $pair = shift(@_);
  my $type = shift(@_);
  my $bonds = shift(@_);
  my @list;

  foreach (sort(keys(%{$bonds}))) {
    my @t =  split("\t");
    foreach(@t) { $_ = $type->{$_}; }
    next if ($pair ne join("\t", arrange(@t)));
    push(@list, $_);
  }
  return @list;
}


sub list_atoms {
  my $pair = shift(@_);
  my $type = shift(@_);
  my $bonds = shift(@_);
  my $reverse = shift(@_);
  my %atoms;

  foreach (list_bonds($pair, $type, $bonds)) {
    foreach (split("\t")) {
      my $t = $type->{$_};
      $t = $reverse->{$t} if (defined($reverse->{$t}));
      $atoms{$t}->{$_} = 1;
    }
  }
  foreach (keys(%atoms)) {
    $atoms{$_} = [sort(keys(%{$atoms{$_}}))];
  }
  return %atoms;
}


sub sort_index_n {
  my @a = @{$a->{n}};
  my @b = @{$b->{n}};

  while (scalar(@a)) {
    return -1 if (@a[0]<@b[0]);
    return 1 if (@a[0]>@b[0]);
    shift(@a); shift(@b);
  }
  return defined($a->{key}) ? $a->{key} cmp $b->{key} : 0;
}


sub create_type_additions {
  my $r = shift(@_);
  my $residue = $r->{residue};
  my $connect = $r->{connect};
  my $charge = $r->{charge};
  my $ntypes = $r->{ntypes};
  my $bonds = shift(@_);
  my $type = shift(@_);
  my $formal = shift(@_);
  my $uncouple = shift(@_);
  my $issues = shift(@_);
  my $debug = $::EMCCharmm::Flag{debug};		# display debug info

  # flow:
  # - create subset of selected bonds from problem increments (issues)
  # - derive candidate selection array split in 
  #   - 0: atoms with multiple neighbors
  #   - 1: atoms with only one neighbor (like hydrogen etc.)
  # - create neighbor list associated with candidates
  # - sort neighbor list on occurence; smallest first
  # - identify possible starting suggestions by looking for atoms with
  #   formal charge neighbors
  
  info("repairing types for residue $residue\n");

  # determine frequency

  my %frequency;
  
  foreach (sort(keys(%{$issues}))) {			# frequency bookkeeping
    foreach (list_bonds($_, $type, $bonds)) {
      foreach (split("\t")) {
	++$frequency{atom}->{data}->{$_};
      }
    }
  }
  foreach ("atom", "type", "ring") {			# sort frequency
    my @data;
    my $flag = $_ ne "atom" ? 1 : 0;
    my $ptr = $frequency{$_}->{data};
    foreach (keys(%{$ptr})) { 
      push(@data, {n => [$ptr->{$_}], key => $_});
      next if ($flag);
      ++$frequency{type}->{data}->{$type->{$_}};
      ++$frequency{ring}->{data}->{$type->{$_}} if (ring($type->{$_}));
    }
    $frequency{$_}->{ordered} = [sort(sort_index_n @data)];
  }
  
  # construct selection

  my (%select, %double, %uncouple);
  my $ftd = $frequency{type}->{data};
  my $nchange = 0;
  
  foreach (keys(%{$uncouple})) {			# create uncouple tree
    my @t = split("\t");
    $uncouple{@t[0]}->{@t[1]} = 1;
    $uncouple{@t[1]}->{@t[0]} = 1;
  }
  foreach (sort(keys(%{$issues}))) {			# register issues
    my $ptr = $select{$_} = [];
    my @bonds = list_bonds($_, $type, $bonds);
    foreach (@bonds) { 
      my (@atoms, @t);
      foreach(split("\t")) { push(@atoms, $_); push(@t, $type->{$_}); }
      next if ($ftd->{@t[0]}==1 && $ftd->{@t[1]}==1);
      next if (defined($uncouple{@t[0]}->{@t[1]}));
      $double{@t[0]} = 1 if (@t[0] eq @t[1]);
      push(@{$ptr}, [@atoms]);
    }
    if (scalar(@{$select{$_}})) { ++$nchange; }
    else { delete($select{$_}); }
  }
  if ($debug) {						# print selection
    tdebug();
    tdebug("nchange", "=", $nchange);
    tdebug();
    tdebug("pairs:");
    foreach (sort(keys(%select))) {
      tdebug("", $_);
      my @b; foreach (@{$select{$_}}) { push(@b, join("\t", @{$_})); }
      @b = sort(@b); while (scalar(@b)) { tdebug("\t", splice(@b,0,3)); }
    }
  }
  if (!$nchange) {
    warning("no increments for residue $residue\n");
    return 0;
  }

  # derive candidate selection array split in 

  my @types = ({}, {});					# find candidate types
  my %allowed;
  
  foreach (sort(keys(%select))) {
    my @type = split("\t");
    foreach (@{$select{$_}}) {
      my $atom = $_;
      for (my $i=0; $i<2; ++$i) {
	my $c = $charge->{$atom->[$i]};
	if ($::EMCCharmm::NConnects{@type[$i]}<2) {
	  @types[1]->{@type[$i]}->{$atom->[$i]} = $c;	# singles
	} else {
	  @types[0]->{@type[$i]}->{$atom->[$i]} = $c;	# candidates
	  $allowed{$atom->[$i]} = $c;
	}
      }
    }
  }
  
  # create neighbor list associated with candidates

  my (@neighbors, @connects, @ordered);			# create neighbors
  my $i = 0;

  foreach (@types) {
    my (%neighs, %cncts);
    my $types = $_;
    foreach (sort(keys(%{$types}))) {
      foreach (sort(keys(%{$types->{$_}}))) {
	my $atom = $_;
	my $t = $type->{$_};
	my @t; 
	foreach (@{$connect->{$_}}) {
	  push(@t, $type->{$_}); $cncts{$atom}->{$_} = 1;
	}
	my $key = join("\t", arrange(@t));
	++$neighs{$t}->{$key}->{n};
	my $ptr = $neighs{$t}->{$key};
	$ptr->{atoms}->{$atom} = $charge->{$atom};
      }
    }
    push(@neighbors, {%neighs});
    push(@connects, {%cncts});
  }

  # sort neighbor list on occurence; smallest first

  foreach (@neighbors) {				# sort neighbors
    my @order;						# smallest first
    my $neighs = $_;
    foreach (keys(%{$neighs})) {
      my @entry;
      my $t0 = $_;
      my $p0 = $neighs->{$t0};
      foreach (sort(keys(%{$p0}))) {
	my $p1 = $p0->{$_};
	push(@entry, {
	    key => $_, n => [$p1->{n}], type => $t0, atoms => $p1->{atoms}});
      }
      next if ($ftd->{$t0}<2);
      push(@order, {
	  key => $t0, n => [-$::EMCCharmm::NConnects{$_}, scalar(@entry)],
	  data => [sort(sort_index_n @entry)]});
    }
    next if (!scalar(@order));
    push(@ordered, [sort(sort_index_n @order)]);
  }
  if ($debug) {
    debug("\n");					# print neighbors
    tdebug("candidates:");
    foreach (@ordered) {
      foreach (@{$_}) {
	my $p0 = $_;
	my $pr = $frequency{ring}->{data};
	tdebug("", $p0->{key}, 
	  $frequency{type}->{data}->{$p0->{key}},
	  defined($pr) ? $pr->{$p0->{key}} : 0
	);
	foreach (@{$p0->{data}}) {
	  my $p1 = $_;
	  my @a; foreach(arrange(keys(%{$p1->{atoms}}))) { push(@a, $_); }
	  tdebug("\t", $p1->{n}->[0], $p1->{key});
	  tdebug("\t\t", @a);
	}
      }
    }
    debug("\n");					# print connections
    tdebug("connects:");
    foreach (@connects) {
      my $connects = $_;
      foreach (sort(keys(%{$connects}))) {
	my @c; foreach (sort(keys(%{$connects->{$_}}))) { push(@c, $_); }
	tdebug("", $_, arrange(@c));
      }
    }
  }

  # identify possible starting suggestions by looking for atoms with
  # formal charge neighbors

  my (%suggest, @suggests);				# starting suggestions

  foreach (sort(keys(%{@connects[0]}))) {		# connections with
    my $a0 = $_;					# formal charge to
    my $t0 = $type->{$a0};
    my $fr = $frequency{ring}->{data};
    next if (defined($fr) && !defined($fr->{$t0}));
    if ($frequency{type}->{data}->{$t0}==1) { $suggest{$a0} = 1; next; }
    foreach (arrange(keys(%{@connects[0]->{$_}}))) {	# candidates
      $suggest{$a0} = 1 if ($formal->{$_});
    }
  }
  foreach (arrange(keys(%suggest))) {
    push(@suggests, [$_]);
  }
  
  my %visited;
  my $iter = 0;
  my $count = $nchange-scalar(keys(%suggest));

  while ($count>0 && $count<$nchange) {			# select neighbors
    my %hash = %suggest;
    foreach (arrange(keys(%hash))) {
      my $key;
      my @select;
      my $a0 = $_;
      my $t0 = $type->{$_};
      my $fr = $frequency{ring}->{data};
      foreach (arrange(keys(%{@connects[0]->{$_}}))) {	# candidates
	next if (defined($fr) && !defined($fr->{$type->{$_}}));
	push(@select, $_) if (defined(@connects[0]->{$_}));
      }
      delete($hash{$_});
      next if (!scalar(@select));
      push(@suggests, [@select]);
      --$nchange;
    }
    last if (++$iter==$nchange);
  }
  undef(%visited);

  if ($debug) {
    debug("\n");					# print suggestions
    tdebug("suggestions:");
    my @s; foreach (@suggests) { push(@s, @{$_}); }
    tdebug("", scalar(@suggests) ? @s : "N/A");
  }

  # retype

  my (%occurence, %retype, %reverse);

  if (scalar(@suggests)) {
    
    # retype suggestions

    foreach (@suggests) {
      my $atom = $_->[0];
      my $t = $type->{$atom};
      if ($frequency{type}->{data}->{$t} == 1) { --$nchange; next; }
      my $typen = new_type($t, \%occurence);

      foreach (@{$_}) { $retype{join("\t", $residue, $_)} = $typen; }
      $::EMCCharmm::Element{$typen} = $::EMCCharmm::Element{$t};
      $reverse{$typen} = $t;
      ++$occurence{$t}->{n};
      --$nchange;
    }
  
  } else {

    # change exactly $nchange types
    # selection criterion: take lowest occurence first

    if (!defined(@ordered[0]->[0]->{data})) {
      debug("\n") if ($debug);
      return 0;
    }

    my (%gather, %ntotal);

    foreach (@{@ordered[0]}) {				# gather by occurence
      my $p0 = $_;
      $ntotal{$p0->{key}} = $p0->{n}->[-1];
      foreach (@{$p0->{data}}) {
	my $p1 = $_;
	my $n = $p1->{n}->[0];
	$gather{$n} = [] if (!defined($gather{$n}));
	push(@{$gather{$n}}, $p1);
      }
    }

    # quick and dirty selection: needs to be recursively generalized
    # currently works for 2 changes only
    
    foreach (sort(keys(%gather))) {			# quick and dirty
      my $ptr = $gather{$_};				# selection of choices
      foreach (@{$gather{$_}}) {
	last if (!$nchange); --$nchange;
	my $choice = $_;				# pick first
	my @atom = sort(keys(%{$choice->{atoms}}));
	my $t = $type->{@atom[0]};

	next if ($ntotal{$t}<2);
	next if ($frequency{type}->{data}->{$t} == 1);
	
	my $typen = new_type($t, \%occurence);
	if ($ntypes->{$t} == scalar(@atom)) {		# as in cgenff TRIB
	  $retype{join("\t", $residue, @atom[0])} = $typen;
	} else {
	  foreach (@atom) { $retype{join("\t", $residue, $_)} = $typen; }
	}
	$::EMCCharmm::Element{$typen} = $::EMCCharmm::Element{$t};
	$reverse{$typen} = $t;
	++$occurence{$t}->{n};
	--$ntotal{$t};
      }
      last if (!$nchange);
    }
  }

  if ($debug) {						# print retypes
    debug("\n");
    tdebug("retype");
    foreach (sort(keys(%retype))) {
      my @a = split("\t");
      tdebug("", @a, $type->{@a[1]}, "->", $retype{$_});
    }
  }

  debug("\n");
  my %rr = get_residue($residue,1,1,$r->{rules},$r->{selection},\%retype);
  if (!assign_increments(\%rr, 1, \%reverse)) {
    warning("no increments for residue $residue\n");
    return 0;
  }
  foreach (keys(%reverse)) { 
    $::EMCCharmm::Reverse{$_} = $reverse{$_};
    delete($::EMCCharmm::Element{$_});
  }
  return 1;
}


sub analyze_increments {
  my @ptr = (\%::EMCCharmm::IncrementTree);
  my $nconnects = \%::EMCCharmm::NConnects;
  my $occurence = \%::EMCCharmm::Occurence;
  my $created = \%::EMCCharmm::TypeCreated;
  my $increment = {};
  my $changed = {};
  my @type;

  info("analyzing increments\n");
  foreach (sort(keys(%{@ptr[0]}))) {			# type 1
    @type[0] = $_;
    @ptr[1] = ${@ptr[0]}{$_};
    foreach (sort(keys(%{@ptr[1]}))) {			# type 2
      @type[1] = $_;
      @ptr[2] = @ptr[1]->{$_};
      my @keys = keys(%{@ptr[2]});
      $increment->{@type[0]}->{@type[1]} = @keys[0];
      next if (scalar(@keys)<2);

      # determine extra types

      my @select;
      foreach (sort(@keys)) {				# order increments
    	@ptr[3] = @ptr[2]->{$_};
	push(@select, {n => [scalar(@{@ptr[3]})], inc => $_});
      }
      @select = reverse(sort(sort_index_n @select));	# max first
      my $orig = shift(@select);			# pop off first
      $increment->{@type[0]}->{@type[1]} = $orig->{inc};
      tdebug("ORIG", @type[0,1], $orig->{inc}, @{$orig->{n}}[0]);
      foreach (@select) {				# retype rest
	my $inc = $_->{inc};
	my %flag; 
	my $typen;

	@ptr[3] = @ptr[2]->{$inc};
	tdebug("RETYPE", @type, $inc, @{$_->{n}}[0]);
	foreach(@{@ptr[3]}) {					
	  my (@data, @all);
	  my $atoms = $_->{atoms};
	  my $resi = $_->{residue};
	  my $change = $_->{change};
	  
	  # determine target

	  foreach (sort(keys(%{$atoms}))) {
	    my $add = {
	      n => [scalar(@{$atoms->{$_}}), -$nconnects->{$_}], type => $_};
	    push(@data, $add) if (defined($change->{$_}));
	    push(@all, $add);
	  }
	  @data = @all if (!scalar(@data));
	  @data = sort(sort_index_n @data);

	  my $target = shift(@data);
	  my $type = $target->{type};
	  my $source = $type;
	  my $reverse = \%::EMCCharmm::Reverse;
	  $type = $reverse->{$type} if (defined($reverse->{$type}));

	  #my $typen;
	  my @select = ();
	  foreach (@{$atoms->{$source}}) {
	    if (defined($changed->{$resi}->{$_})) {
	      $typen = $changed->{$resi}->{$_};
	    } else {
	      push(@select, $_);
	    }
	  }
	   
	  if (scalar(@select)) {
	    $flag{$type} = 1;
	    if ($typen eq "") {
	      $typen = create_mass_entry($type, $resi, \@select);
	    }
	    $created->{$typen} = $type;
	    foreach (@select) { $changed->{$resi}->{$_} = $typen; }
	  }
	  tdebug("", 
	    $resi, $type, $typen, $nconnects->{$type}, @{$atoms->{$source}});

	  # assign new increment

	  my @t = @type[0,1];
	  foreach (@t) { next if ($_ ne $type); $_ = $typen; last; }
	  @t = arrange(@t); $increment->{@t[0]}->{@t[1]} = $inc;
	}
	foreach (keys(%flag)) { ++$occurence->{$_}->{n}; }
      }
      debug("\n");
    }
  }

  # => EMC takes care of cross terms: uses pair equivalence when increment
  # does not exist

  # transcribe

  undef($::EMCCharmm::Increments);
  foreach (keys(%{$increment})) {
    my $type = $_;
    my $ptr = $increment->{$_};
    foreach (keys(%{$ptr})) {
      my $key = join("\t", arrange($type, $_));
      $::EMCCharmm::Increments{$key} = $ptr->{$_};
    }
  }
  foreach (keys(%::EMCCharmm::IncrementsRead)) {
    next if (defined($::EMCCharmm::Increments{$_}));
    $::EMCCharmm::Increments{$_} = $::EMCCharmm::IncrementsRead{$_}->[1];
  }
}


sub create_replicas {
  my $mode = shift(@_);
  my $replica = \%::EMCCharmm::Replica;
  my $nconnects = \%::EMCCharmm::NConnects;

  foreach (sort(keys(%{$replica}))) {
    my $type = $replica->{$_}->{src};
    if ($mode==0) {
      my $typen = create_mass_entry($type, "", [], $_);
      $replica->{$_}->{dest} = $typen;
      info("creating replica '$typen' from '$type'\n");
    } elsif(defined($nconnects->{$type})) {
      my $typen = $replica->{$_}->{dest};
      $nconnects->{$typen} = $nconnects->{$type};
    }
  }
}


sub create_increments {
  my $fgroup = 1;
  my $depth = 1;
  my %selection;
  my %rules;
  my @list;

  info("creating increments\n");
  if ($::EMCCharmm::Flag{include}) {
    @list = @{$::EMCCharmm::Residue{includes}};
  } else {
    @list = @{$::EMCCharmm::Residue{residues}};
  }
  @list = sort(@list) if ($::EMCCharmm::Flag{sort});
  foreach (@list) { 
    @{$selection{$_}} = ();
  }
  foreach (@list) {					# loop over selection
    my %r = get_residue($_, $depth, $fgroup, \%rules, \@{$selection{$_}});
    next if (!defined($r{residue}));
    assign_increments(\%r, 0);
    return 0 if ($r{stop});
  }
  analyze_increments();
  
  return 1 if (!%::EMCCharmm::Occurence);
  my @data = sort(keys(%::EMCCharmm::TypeCreated));
  my $n = scalar(@data);
  info("created $n new type".($n==1 ? "" : "s")."\n");
  while (scalar(@data)) {
    info(join("\t", "", splice(@data,0,8))."\n");
  }
  return 1;
}


# create single rule from selection

sub create_residue_rule {
  my %r = get_residue(@_);

  if (!(defined($r{residue})) && 
      (!defined($r{selection}) || scalar(@{$r{selection}}))) { return; }

  tdebug("rules:");
  foreach (sort(@{$r{selection}})) {			# update rules
    next if (substr($_,0,1) =~ m/\+|\-|\*/);
    my (%visited, $key, %rules);
    my $type = $r{type}->{$_};
    my $charge = $::EMCCharmm::Flag{increment} && 
		 %{$r{bond}} ? 0 : $r{charge}->{$_};
    my $e = $::EMCCharmm::Element{$type};
    my $rule = create_rule(
	$r{residue}, $_,
       	$r{fgroup} ? $r{group}->{$_} : -1, $::EMCCharmm::Element{$type}, 
	$r{group}, $r{type}, $r{connect}, $r{bond}, $r{charge}, 
	\%visited, \%rules, $r{depth}, 0, "", "");
    if (!defined($r{connect}->{$_})) {
      $rule = element($r{residue},
	  $_, $r{connect}, $r{bond}, $r{type}, $r{charge}->{$_});
    }
    debug("\t$_\t$rule\n");

    $key = join("\t", 
      ucfirst($::EMCCharmm::Element{$type}), $rule, $type, $charge);
    #next if (defined($rules->{$key}));
    $r{rules}->{$key} .= ":" if (defined($r{rules}->{$key}));
    $r{rules}->{$key} .= "$r{index}\t$_";
  }
  debug("\n");
}


sub assign_rule {
  my $counter = shift(@_);
  my $key = shift(@_);
  my $id = shift(@_);
  my $value = shift(@_);
  my $current = shift(@_);
  my $depth = shift(@_);
  my $depths = shift(@_);
  my @arg = $key ne "" ? split("\t", $::EMCCharmm::Rules{$key}) : ();

  if ($counter==1) {
    if ($current eq "" && $depths->{$id}<=$depth) {
      $::EMCCharmm::Rules{$key} = $value; return 0;
    }
  } elsif (scalar(@arg)) {
    if (@arg[0] eq (split("\t", $value))[0] && $depths->{$id}<=$depth) {
      $::EMCCharmm::Replace{$value} = $key;
      delete $::EMCCharmm::Rules{$key};
    }
  }
  return $key ne "";
}


sub work_rules {
  my $depth = shift(@_);
  my $rules = shift(@_);
  my $depths = shift(@_);
  my $last_depths = shift(@_);
  my $last_rules = shift(@_);

  foreach (keys(%$rules))
  {
    my $id = $_;
    my $current = join("\t", (split("\t"))[2,1,3]);

    foreach (split(":", $rules->{$id})) {
      my $key = $_;
      my $last = join("\t", (split("\t", $last_rules->{$key}))[2,1,3]);
      
      $depths->{$key} = $depth if ($depths->{$key}<$depth && $current ne $last);
      $last_depths->{$key} = $depths->{$key};
      $last_rules->{$key} = $id;
    }
  }
}


sub add_selection {
  my $selection = shift(@_);
  my $values = shift(@_);

  foreach (@$values) {
    my @arg = split("\t");
    push(@{$selection->{@arg[0]}}, @arg[1]);
  }
}


# Create rules and increments from residues

sub get_rules {
  my %selection = ();
  my ($depth, %depths, %last_depths, %last_rules);
  my ($residue, $atom, $element, $type, $charge, $rule);
  my (@rules, @last_rules);
  my $ntrials = 2;
  my @values = ();
  my $max_depth = 0;

  info("creating rules\n");
  foreach (sort(keys(%{$::EMCCharmm::Residue{index}}))) {
    @{$selection{$_}} = ();
  }
  foreach (sort(keys(%{$::EMCCharmm::Residue{patch}}))) {
    @{$selection{$_}} = ();
  }
  while ($ntrials--) {					# iterative rule scheme
    $depth = 0;
    debug("\n");
    info("%s\n", $ntrials ? "group delimited" : "free", "scan");
    debug("\n");
    while (scalar(keys(%selection)) && $depth++<$::EMCCharmm::FFDepthMax) {
      my ($key, $id, $value, $current, $result);
      my %rules = ();
      my $counter = 0;
      my $last = "";
      my @arg = ();

      info("scanning depth = $depth\n");
      debug("\n"); debug("redo:\n"); debug("\n");
      foreach (sort(keys(%selection))) {		# rules from selection
	#debug("$_\t@{$selection{$_}}\n");
	create_residue_rule($_, $depth, $ntrials, \%rules, \@{$selection{$_}});
      }
      work_rules($depth, \%rules, \%depths, \%last_depths, \%last_rules);

      @rules = ();
      %selection = ();
      debug("\n"); debug("rules:\n"); debug("\n");
      foreach (sort(keys(%rules))) {			# assign rules 
	($element, $rule, $type, $charge) = split("\t", $_);
	if ($last ne $rule) {
	  add_selection(\%selection, \@values) if (assign_rule(
	      $counter, $key, $id, $value, $current, $depth, \%depths));
	  $counter = 0;
	} elsif ($last ne "") {
	  add_selection(\%selection, \@values) if (assign_rule(
	      2, $key, $id, $value, $current, $depth, \%depths));
	}
	@values = ();
	$key = join("\t", $charge, $rule);
	$current = $::EMCCharmm::Rules{$key};
	foreach (reverse(split(":", $rules{$_}))) {
	  ($residue, $atom) = split("\t");
	  push(@values, $id = join("\t", $residue, $atom));
	  debug(join("\t",
	      $depths{$id}, $type, $residue, $atom, $charge, $rule), "\n");
	}
	$value = join("\t", $type, $residue, $atom);
	push(@rules, $id);
	$last = $rule;
	++$counter;
      }
      if (join("\t", @rules) ne join("\t", @last_rules)) {
       	$::EMCCharmm::FFDepth = $depth if ($::EMCCharmm::FFDepth<$depth);
      }
      @last_rules = @rules;
      debug("\n");
      debug("set max depth = $::EMCCharmm::FFDepth\n");
      debug("\n");
      push(@{$selection{$residue}}, $atom) if (assign_rule(
	  $counter, $key, $id, $value, $current, $depth, \%depths));
    }
    $ntrials = 0 if (join("\t", @rules) eq "");
    $max_depth = $depth if ($max_depth<$depth);
  }
  $::EMCCharmm::FFDepth = $max_depth if ($::EMCCharmm::FFDepth>$max_depth);
  if ($::EMCCharmm::FFDepth < $::EMCCharmm::FFDepthRead) {
    $::EMCCharmm::FFDepth = $::EMCCharmm::FFDepthRead;
  }
  @rules = ();

  my %rules = ();					# check for doubles
  my %dpths = ();
  my $doubles = 0;
  
  foreach (keys(%::EMCCharmm::Rules)) {			# check for doubles
    ($charge, $rule) = split("\t");
    ($type, $residue, $atom) = split("\t", $::EMCCharmm::Rules{$_});
    my $dpth = $depths{join("\t", $residue, $atom)};
    if (push(@{$rules{$rule}}, $charge)<2) { $dpths{$rule} = $dpth; next; }
    if ($dpths{$rule}>$dpth || $dpths{$rule} eq "") {
      $dpths{$rule} = $dpth if ($dpth ne "");
    }
    $doubles = 1;
  }
  if ($doubles) {
    debug("doubles:\n"); debug("\n");
    foreach (sort(keys(%rules))) {
      next if (scalar(@{$rules{$rule = $_}})<2);
      debug(join("\t", @{$rules{$rule}}, $rule, "\n"));
      $depth = $dpths{$rule};
      my $success = 0;
      my %replace = ();
      for (; $depth<$::EMCCharmm::FFDepth && !$success; ++$depth) {
	my @last;
	foreach (@{$rules{$rule}}) {
	  my $key = join("\t", $charge = $_, $rule);
	  ($type, $residue, $atom) = split("\t", $::EMCCharmm::Rules{$key});
	  next if ($residue eq "-");
	  my %rls = ();
	  my %select = ();
	  @{$select{$residue}} = ($atom);
	  create_residue_rule(
	      $residue, $depth, 0, \%rls, \@{$select{$residue}});
	  my @r = split("\t", (keys(%rls))[0]);
	  if (!scalar(@last)) { @last = (@r[-1, 1], $key); next; }
	  last if (@last[1] eq @r[1]);
	  @{$replace{@last[2]}} = @last[0,1] if (@last[2] ne "");
	  @{$replace{$key}} = @last = @r[-1, 1];
	  $success = 1;
	}
      }
      foreach (@{$rules{$rule}}) {
	my $key = join("\t", $charge = $_, $rule);
	($type, $residue, $atom) = split("\t", $::EMCCharmm::Rules{$key});
	if ($residue eq "-" || !defined($replace{$key})) {
	  debug(join("\t", $::EMCCharmm::Rules{$key}, "\n")); next;
	}
	my @r = @{$replace{$key}};
	debug(join("\t", $::EMCCharmm::Rules{$key}, @r, "\n"));
	$::EMCCharmm::Rules{join("\t", $charge, @r[-1])} = $::EMCCharmm::Rules{$key};
	delete($::EMCCharmm::Rules{$key});
      }
      debug("\n");
    }
  }

  my @replace;			# @::EMCCharmm::Rules from %::EMCCharmm::Rules
  
  foreach (keys(%::EMCCharmm::Rules)) {	
    my $value = $::EMCCharmm::Rules{$_};
    my $replace = $::EMCCharmm::Replace{$value};
    my @arg = (split("\t", $replace eq "" ? $_ : $replace),split("\t", $value));
    my $element = $::EMCCharmm::Element{@arg[2]}; $element =~ s/\[|\]//g;
    push(@rules, join("\t", ucfirst($element), @arg[2,1,0,3,4]));
    push(@replace, join("\t", $value, $replace, "\n")) if ($replace ne "");
  }
  
  if ($::EMCCharmm::Flag{debug} && scalar(@replace)) {
    debug("replacements:\n"); debug("\n");
    foreach (sort(@replace)) { debug($_); }
    debug("\n");
  }

  @::EMCCharmm::Types = ();
  my %test = ();
  
  foreach (@::EMCCharmm::Rules = sort(@rules)) {
    my $type = (split("\t"))[1]; next if ($test{$type});
    push(@::EMCCharmm::Types, $type); $test{$type} = 1;
  }
 
  if (scalar(keys(%selection))) {
    warning("maximum recursion depth exceeded\n");
    foreach (sort(keys(%selection))) {
      warning("missing: $_\t[@{$selection{$_}}]\n");
    }
  }
}


# parameter analysis

sub check_types {
  my $flag = 1;
  my @missing = ();

  foreach (@_) {
    next if ($_ =~ m/\*/ || $_ =~ m/\?/);
    return 0 if (!defined($::EMCCharmm::NConnects{$_}));
    next if (defined($::EMCCharmm::Masses{$_}));
    warning("ignoring [$_\t]: no mass entry\n");
    $flag = 0;
  }
  return $flag;
}


sub parameters_combine {
  my $set = shift(@_);
  my $title = shift(@_);
  my @keys = sort(keys(%{$set}));
  my %combined;
  my %n;
  
  foreach (@keys) {
    my @arg = split("\t");
    my @data = split("\t", $set->{$_});
   
    foreach (@arg) {
      $_ = $::EMCCharmm::Combine{$_} if (defined($::EMCCharmm::Combine{$_}));
    }
    my $key = join("\t", @arg);
    if (!defined($n{$key})) {
      $combined{$key} = [@data];
    } else {
      my $ptr = $combined{$key};
      for (my $i=0; $i<scalar(@data); ++$i) {
	@{$ptr}[$i] += @data[$i];
      }
    }
    ++$n{$key};
  }
  foreach (keys(%n)) {
    my $nn = $n{$_};
    foreach (@{$combined{$_}}) { $_ /= $nn; }
    $combined{$_} = join("\t", @{$combined{$_}});
  }
  %{$set} = %combined;
}


sub parameters_check {
  
  # for creation of precendences - under construction
  
  my $set = shift(@_);
  my $title = shift(@_);
  my $ptr = shift(@_);
  my @keys = sort(keys(%{$set}));
  my $equivalence;
  my %check;

  print("---\n\n");
  print("$title\n\n");
  foreach (sort(keys(%{$set}))) {
    my $params = ${$set}{$_};
    $check{$params} = [] if (!defined($check{$params}));
    push(@{$check{$params}}, $_);
  }
  foreach (sort(keys(%check))) {
    my @args = @{$check{$_}};
    foreach (@args) {
      my @arg1 = split("\t");
      foreach (@args) {
	my @arg2 = split("\t");
	my $i; for ($i=0; $i<scalar(@arg1); ++$i) {
	  next if (substr($arg1[$i],0,2) ne substr($arg2[$i],0,2));
	  ${${$equivalence}{$arg1[$i]}}{$arg2[$i]} = 1;
	}
      }
    }
    print("$_:\n  ", join("\n  ", sort(@{$check{$_}})), "\n");
  }
  print("\n");
  print("Equivalence $title\n\n");
  foreach (sort(keys(%{$equivalence}))) {
    print("$_:\t", join("\t", sort(keys(%{${$equivalence}{$_}}))), "\n");
  }
  print("\n");
}


# Pre-analysis

sub analyze() {
  my %equivalences;
  my @set = (
    [\%::EMCCharmm::Nonbond, "Nonbond"],
    [\%::EMCCharmm::Increments, "Increments"],
    [\%::EMCCharmm::Bond, "Bond"],
    [\%::EMCCharmm::Bond, "BondAuto"],
    [\%::EMCCharmm::Angle, "Angle"],
    [\%::EMCCharmm::Angle, "AngleAuto"],
    [\%::EMCCharmm::Torsion, "Torsion"],
    [\%::EMCCharmm::TorsionAuto, "TorsionAuto"], 
    [\%::EMCCharmm::Improper, "Improper"],
    [\%::EMCCharmm::ImproperAuto, "ImproperAuto"]);

  my $i; for ($i=0; $i<scalar(@set); ++$i) {
    parameters_combine(@{$set[$i]});
    #parameters_check(@{$set[$i]}, \%equivalences);
  }
  
  # scrub residue lists

  foreach(@{$::EMCCharmm::Residue{lists}}) {
    my @list;
    my %check;
    my $ptr = $::EMCCharmm::Residue{$_};

    foreach (reverse(@{$ptr})) {
      push(@list, $_) if (!defined($check{$_}));
      $check{$_} = 1;
    }
    @{$ptr} = reverse(@list);
  }
}


# settings processing

sub set_precedences {
  my @precedences;
  my $first = shift(@::EMCCharmm::Precedences);
  my $last = pop(@::EMCCharmm::Precedences);

  $first = "(?" if ($first eq "");
  $last = ")" if ($last eq "");
  foreach (@::EMCCharmm::Types) {
    my $flag = 0;
    my $type = $_;
    foreach (@::EMCCharmm::Precedences) {
      my @arg = split(" ");
      foreach (@arg) {
	$_ =~ s/\(//g;
	$_ =~ s/\)//g;
	$flag = ($_ eq $type ? 1 : 0);
	last if ($flag); 
      }
      last if ($flag);
    }
    push(@precedences, "  ($_)") if (!$flag);
  }
  @precedences = sort({
      my $c = substr($a,0,4);
      my $d = substr($b,0,4);
      $c lt $d ? -1 : $c gt $d ? 1 : 0 } @precedences);
  push(@::EMCCharmm::Precedences, @precedences);
  unshift(@::EMCCharmm::Precedences, $first);
  push(@::EMCCharmm::Precedences, $last);
}


sub set_equivalences {
  my ($index, $type, $typen, $element);

  foreach (keys(%::EMCCharmm::Masses)) {
    my $type = $_;
    my $typen = $_;
    if (!defined($::EMCCharmm::Equivalence{$typen})) {
      my @data = ($type, $type, $type, $type, $type, $type);
      $::EMCCharmm::Equivalence{$typen} = [@data];
      $::EMCCharmm::EquivalenceAuto{$typen} = [@data];
    }
  }
}


sub set_extras {
  my (@arg, @extra, %torsion, %torsion_double, %offset, $key);
  my ($index, $basis, $type, $charge, $itorsion);
  my ($id, $type, $dummy, $sigma, $epsilon, $comment);
  my $ntd = $::EMCCharmm::NTorsionDoubles;

  foreach (keys(%::EMCCharmm::Extras)) {
    @extra = split("\t", $::EMCCharmm::Extras{$_});
    ($index, $basis, $type, $charge) = splice(@extra, 0, 4);
    if (!defined($::EMCCharmm::Rules{$index})) {
      error("no rule exists for extra index $index\n");
    }
    if (!defined($::EMCCharmm::Masses{$type})) {		# masses
      @arg = @{$::EMCCharmm::Masses{$basis}->{data}};
      @arg[0] = $type;
      $::EMCCharmm::Masses{$type} = {
	data => [@arg], order => $::EMCCharmm::MassOrder};
      $::EMCCharmm::Element{$type} = $::EMCCharmm::Element{$basis};
    }
    elsif (!defined($::EMCCharmm::Element{$type})) {
      $::EMCCharmm::Element{$type} = $::EMCCharmm::Element{$basis};
    }
    if (!defined($::EMCCharmm::Equivalence{$type})) {		# equivalence
      @arg = @{$::EMCCharmm::EquivalenceAuto{$basis}};
      $::EMCCharmm::EquivalenceAuto{$type} = [@arg];
      @arg = @{$::EMCCharmm::Equivalence{$basis}}; @arg[3] = $type;
      $::EMCCharmm::Equivalence{$type} = [@arg];
    }
    @arg = split("\t", $::EMCCharmm::Index{int($index)});	# index
    $charge = @arg[4] if ($charge eq "" || $charge eq "-");
    @arg[3,4] = ($type, $charge);
    $::EMCCharmm::Index{$index} = join("\t", @arg);
    $::EMCCharmm::Type{$type} = 1;
    foreach (@extra) {						# torsions
      ($itorsion, $id) = split(":");
      foreach (split("\t", $::EMCCharmm::TorsionIndex{int($itorsion)})) {
	next if (!defined($::EMCCharmm::TorsionKey{$_}));	# update index
	@arg = split("\t", $key  = $::EMCCharmm::TorsionKey{$_});
	foreach (@arg[0,1,2,3]) {
	  $_ = $type if ($_ eq $basis);
	}
	$torsion{join("\t", @arg)} = $::EMCCharmm::Torsion{$key};
	my @a = split("\t", $::EMCCharmm::Torsion{$key});
	$::EMCCharmm::Torsion{$key} = "";
      }
      if (defined($::EMCCharmm::TorsionDoubleIndex{int($itorsion)})) {
	foreach (split("\t", $::EMCCharmm::TorsionDoubleIndex{int($itorsion)})) {
	  next if ($id && ($itorsion eq $_));
	  next if (!defined($::EMCCharmm::TorsionDouble{$_}));
	  @arg = split("\t", $::EMCCharmm::TorsionDouble{$_});
	  foreach (@arg[0,1,2,3]) {
	    $_ = $type if ($_ eq $basis);
	  }
	  $::EMCCharmm::TorsionDouble{$_} = join("\t", @arg);
	  $torsion_double{@arg[-1]} = 1;
	}
      }
      next if (!$id);
      if (defined($::EMCCharmm::TorsionDouble{$itorsion})) {	# set double
	@arg = split("\t", $::EMCCharmm::TorsionDouble{$itorsion});
	foreach (@arg[0,1,2,3]) { 
	  if (($_ eq $basis) && $id--) { $_ = $type; last;  }
	}
	error("invalid occurence".
	  " in extra index $index for torsion $itorsion\n") if ($id);
	$::EMCCharmm::TorsionDouble{$itorsion} = join("\t", @arg);
	$torsion_double{@arg[-1]} = 1;
      }
    }
  }
  foreach (keys(%::EMCCharmm::Torsion)) {
    $torsion{$_} = $::EMCCharmm::Torsion{$_} if ($::EMCCharmm::Torsion{$_} ne "");
  }
  foreach (keys(%torsion_double)) {
    @arg = split("\t", $::EMCCharmm::TorsionDouble{$_});
    @arg[0,1,2,3] = arrange(@arg[0,1,2,3]);
    my $key = join("\t", splice(@arg, 0, 4));
    $torsion{$key} = join("\t", @arg);
    --$::EMCCharmm::NTorsionDoubles;
  }
  %::EMCCharmm::Torsion = %torsion;
  if ($::EMCCharmm::NBondDoubles) {
    warning("bond doubles exist\n");
  }
  if ($N::EMCCharmm::Angle::EMCCharmm::Doubles) {
    warning("angle doubles exist\n");
  }
  if ($::EMCCharmm::NTorsionDoubles) {
    my $message = $ntd==$::EMCCharmm::NTorsionDoubles ? 
	"($ntd in total)" : "(reduced from $ntd to $::EMCCharmm::NTorsionDoubles)";
    warning("torsion doubles exist $message\n");
  }
}


sub set_missing {
  info("create missing impropers\n");
  foreach (
      [\%::EMCCharmm::ImproperAuto, \%::EMCCharmm::ImproperReverseAuto], 
      [\%::EMCCharmm::Improper, \%::EMCCharmm::ImproperReverse]) {
    my $ptr = $_;
    foreach (sort(keys(%{@{$ptr}[0]}))) {
      my @arg = split("\t");
      next if (!defined($::EMCCharmm::NConnects{@arg[-1]}));
      next if ($::EMCCharmm::NConnects{@arg[-1]}!=3);
      my $key = join("\t", reverse(@arg));
      next if (defined(${@{$ptr}[0]}{$key}));
      info("\t$_\t->\t$key\n");
      ${@{$ptr}[0]}{$key} = ${@{$ptr}[0]}{$_};
      ${@{$ptr}[1]}{$key} = ${@{$ptr}[1]}{$_};
    }
  }
}


# type reduction (after parameter read and increments)

sub reassign_mass_order {
  my @tmp;
  my $n = 0;
  my @masses;
  my $mass = \%::EMCCharmm::Masses;
  
  foreach (sort(keys(%{$mass}))) {
    if (!@masses[$mass->{$_}->{order}]) {
      @masses[$mass->{$_}->{order}] = [];
    }
    push(@{@masses[$mass->{$_}->{order}]}, $_);
  }
  foreach (@masses) {
    push(@tmp, $_) if ($_);
  }
  foreach(@tmp) {
    foreach(@{$_}) {
      $mass->{$_}->{order} = $n;
    }
    ++$n;
  }
}


sub order_torsion {
  foreach (\%::EMCCharmm::TorsionAuto, \%::EMCCharmm::Torsion) {
    my $torsion = $_;
    foreach (sort(keys(%{$torsion}))) {
      my ($i, @p, @t);
      my $key = $_;
      my @arg = split("\t", $torsion->{$key});
      for ($i=0; $i<scalar(@arg); $i+=3) {
	push(@p, join("\t", @arg[$i+1,$i,$i+2]));
      }
      foreach (sort(@p)) { 
	push(@t, (split("\t"))[1,0,2]);
      }
      $torsion->{$key} = join("\t", @t);
    }
  }
}


sub mass_order_lists {
  my %list;
  my $mass = \%::EMCCharmm::Masses;

  foreach (sort(keys(%{$mass}))) {
    my $n = $mass->{$_}->{order};
    my $m = round(@{$mass->{$_}->{data}}[0], 0.001);
    $list{$m} = [] if (!defined($list{$m}));
    @{$list{$m}}[$n] = [] if (!@{$list{$m}}[$n]);
    push(@{@{$list{$m}}[$n]}, $_);
  }
  foreach (sort({$a <=> $b} keys(%list))) {
    my $n = 0;
    tprint();
    tprint($_);
    foreach (@{$list{$_}}) {
      tprint("", $n, @{$_}) if ($_);
      ++$n;
    }
  }
  print("\n");
}


sub reduce_bonds {
  my %list;
  my $bond = \%::EMCCharmm::Bond;
  my $mass = \%::EMCCharmm::Masses;

  foreach (sort(keys(%{$bond}))) {
    my $n = 0; foreach (split("\t")) { 
      $n = $mass->{$_}->{order} if ($n<$mass->{$_}->{order});
    }
    my $m = $bond->{$_};
    $list{$m} = [] if (!defined($list{$m}));
    @{$list{$m}}[$n] = [] if (!@{$list{$m}}[$n]);
    push(@{@{$list{$m}}[$n]}, $_);
  }
  foreach (sort(keys(%list))) {
    my $n = 0;
    tprint();
    tprint($_);
    foreach (@{$list{$_}}) {
      tprint("", $n, @{$_}) if ($_);
      ++$n;
    }
  }
  print("\n");
}


sub reduce_types {
  reassign_mass_order();
  mass_order_lists();
  reduce_bonds();
  order_torsion();
}


# output parameter and topology file

sub print_header {
  my $stream = shift(@_);
  my $text = shift(@_);
  my $definition = shift(@_);

  printf($stream "#EMC/FIELD/CHARMM/$::EMCCharmm::FFVersion
#
#  CHARMM $text using $::EMCCharmm::ProjectName force field files, 
#  converted by $::EMCCharmm::script $::EMCCharmm::version, $::EMCCharmm::date by $::EMCCharmm::author,
#  to be used in conjuction with EMC v$::EMCCharmm::EMCVersion or higher,
#  and created on $::EMCCharmm::Date.
#\n");

  return if (!$definition);
  printf($stream "\n# Force field definition\n\n");
  printf($stream "ITEM\tDEFINE\n\n");
  printf($stream "FFMODE\t\t$::EMCCharmm::FFMode\n");
  printf($stream "FFTYPE\t\tATOMISTIC\n");
  printf($stream "FFAPPLY\t\t$::EMCCharmm::FFApply\n");
  printf($stream "FFINDEX\t\t$::EMCCharmm::FFIndex\n");
  printf($stream "FFDEPTH\t\t$::EMCCharmm::FFDepth\n");
  printf($stream "FFCHARGE\t$::EMCCharmm::FFCharge\n");
  printf($stream "VERSION\t\t$::EMCCharmm::FFVersion\n");
  printf($stream "CREATED\t\t".`date +"%b %Y"`);
  printf($stream "LENGTH\t\tANGSTROM\n");
  printf($stream "ENERGY\t\tKCAL/MOL\n");
  printf($stream "DENSITY\t\tG/CC\n");
  printf($stream "MIX\t\tBERTHELOT\n");
  printf($stream "NBONDED\t\t3\n");
  printf($stream "INNER\t\t$::EMCCharmm::Inner\n");
  printf($stream "CUTOFF\t\t$::EMCCharmm::Cutoff\n");
  printf($stream "ANGLE\t\tERROR\n");
  printf($stream "PAIR14\t\tINCLUDE\n");
  printf($stream "TORSION\t\tERROR\n");
  printf($stream "GROUP\t\tIGNORE\n");
  printf($stream "RULE\t\t$::EMCCharmm::FFRule\n");
  printf($stream "SHAKE\t\t$::EMCCharmm::FFShake\n");
  printf($stream "\nITEM\tEND\n");
}


# output parameter file

sub print_references {
  my $stream = shift(@_);

  return if (!scalar(@::EMCCharmm::References));

  my $n = scalar(@::EMCCharmm::References);

  info("writing $n reference".($n==1 ? "" : "s")."\n");
  printf($stream "\n# Literature references\n\n");
  printf($stream "ITEM\tREFERENCES\n\n");
  printf($stream "# year\tvolume\tpage\tjournal\n\n");
  foreach(sort(@::EMCCharmm::References)) {
    my @arg = split("\t");
    printf($stream "%s\n", join("\t", @arg[0,2,3,1]));
  }
  printf($stream "\nITEM\tEND\n");
}


sub print_masses {
  my $stream = shift(@_);

  return if (!scalar(keys(%::EMCCharmm::Masses)));

  my $n = 0;
  foreach(keys(%::EMCCharmm::Masses)) {
    ++$n if (defined($::EMCCharmm::NConnects{$_}));
  }

  info("writing $n mass".($n==1 ? "" : "es")."\n");
  printf($stream "\n# Masses\n\n");
  printf($stream "ITEM\tMASS\n\n");
  printf($stream "# type\tmass\telement\tncons\tcharge\tindex\tcomment\n\n");
  foreach(sort(keys(%::EMCCharmm::Masses))) {
    next if (!defined($::EMCCharmm::NConnects{$_}));
    my @arg = @{$::EMCCharmm::Masses{$_}->{data}};
    @arg[1] = ucfirst(@arg[1]);
    @arg[0] = get_mass(@arg[1]) if ($::EMCCharmm::Flag{mass});
    printf($stream "$_\t%s\n",
      join("\t", @arg[0,1], 
	$::EMCCharmm::NConnects{$_}, $::EMCCharmm::Charge{$_}, @arg[2,3]));
  }
  printf($stream "\nITEM\tEND\n");
}


sub print_cmap {
  my $stream = shift(@_);

  return if (!scalar(keys(%::EMCCharmm::CMAP)));

  my $n = 0;
  foreach(keys(%::EMCCharmm::CMAP)) {
    ++$n if (check_types(split("\t")));
  }

  printf($stream "\n# CMAP parameters\n\n");
  printf($stream "ITEM\tCMAP\n\n");
  printf($stream "# type1\ttype2\ttype3\ttype4\ttype5\tdim\ttable\n");
  #printf($stream "# k11\t[...]\n");
  #printf($stream "# [...]\n");
  foreach (sort(keys(%::EMCCharmm::CMAP))) {
    my $key = $_;
    my $index = join("\t", (split("\t", $key))[0,1,2,3,-1]);
    next if (!check_types(split("\t")));
    my @data = @{${$::EMCCharmm::CMAP{$key}}{data}};
    printf($stream "\n$index\t%d\n", ${$::EMCCharmm::CMAP{$key}}{n});
    while (scalar(@data)) {
      my $i = 0;
      my $m = 5;
      my $n = scalar(@data);
      foreach (splice(@data,0,$n<$m ? $n : $m)) {
	printf($stream "%s%10.6f", $i++ ? "\t" : "", $_);
      }
      printf($stream "\n");
    }
  }
  printf($stream "\nITEM\tEND\n");
}


sub print_parameters_section {
  my $stream = shift(@_);
  my $items = shift(@_);
  my $select = lc(shift(@_));
  my $check = shift(@_);
  my $Select = ucfirst($select);
  my $SELECT = uc($select.($select eq "template" ? "s" : ""));
  my $ptr = $items->{$select};
  my @header = @{$ptr->{header}};
  my $parameters = defined($ptr->{parameters}) ? 1 : 0;
  my $increments = $select eq "increment" ? 1 : 0;
  my $array = defined($ptr->{array}) ? 1 : 0;
  my $n = 0;

  foreach ("auto", "spec") {
    next if (!defined($ptr->{$_}));
    foreach (keys(%{@{$ptr->{$_}}[0]})) {
      ++$n if ($check ? check_types(split("\t")) : 1);
    }
  }

  info("writing $n ".$select.($n==1 ? "" : "s")."\n");

  foreach ("auto", "spec") {
    next if (!defined($ptr->{$_}));

    my $auto = $_ eq "auto" ? 1 : 0;
    my @data = @{$ptr->{$_}};

    next if (!scalar(keys(%{@data[0]})));

    printf($stream "\n# %s", $parameters ?
      "$Select ".($auto ? "wildcard " : "")."parameters\n\n" :
      ($auto ? "Auto $select" : "$Select")."s\n\n"
    );
    printf($stream "%s", "ITEM\t$SELECT".($auto ? "_AUTO" : "")."\n\n");
    printf($stream "%s", "# ".join("\t", @header)."\n\n");
    foreach (sort(keys(%{@data[0]}))) {
      next if ($check ? !check_types(split("\t")) : 0);
      my $key = $_;
      printf($stream "$key");
      foreach (@data) {
	next if (!defined($_->{$key}));
       	printf($stream "%s",
	  "\t".($array ? join("\t", @{$_->{$key}}) : $_->{$key}));
      }
      printf($stream "\n");
    }
    printf($stream "\nITEM\tEND\n");
  }
}


sub print_parameters {
  my %items = (
    equivalence => {
      header => ["type", "pair", "incr", "bond", "angle", "torsion", "impr"],
      auto => [\%::EMCCharmm::EquivalenceAuto],
      spec => [\%::EMCCharmm::Equivalence],
      array => 1
    },
    increment => {
      header => ["type1", "type2", "delta12", "delta21"],
      spec => [\%::EMCCharmm::Increments],
      parameters => 1
    },
    nonbond => {
      header => ["type1", "type2", "rmin/2", "ewell", "rmin14/2", "ewell14"],
      spec => [\%::EMCCharmm::Nonbond],
      parameters => 1
    },
    bond => {
      header => ["type1", "type2", "k", "l0"],
      auto => [\%::EMCCharmm::BondAuto],
      spec => [\%::EMCCharmm::Bond],
      parameters => 1
    },
    angle => {
      header => ["type1", "type2", "type3", "k", "theta0", "kub", "s0"],
      auto => [\%::EMCCharmm::AngleAuto],
      spec => [\%::EMCCharmm::Angle],
      parameters => 1
    },
    torsion => {
      header => ["type1", "type2", "type3", "type4", "k", "n", "delta", "[...]"],
      auto => [\%::EMCCharmm::TorsionAuto],
      spec => [\%::EMCCharmm::Torsion],
      parameters => 1
    },
    improper => {
      header => ["type1", "type2", "type3", "type4", "k", "psi", "reverse"],
      auto => [\%::EMCCharmm::ImproperAuto, \%::EMCCharmm::ImproperReverseAuto],
      spec => [\%::EMCCharmm::Improper, \%::EMCCharmm::ImproperReverse],
      parameters => 1
    },
  );

  print_header($::EMCCharmm::OUT_PARMS, "interaction parameters", 1);
  print_references($::EMCCharmm::OUT_PARMS);
  print_masses($::EMCCharmm::OUT_PARMS);
  foreach ("equivalence",
    "increment", "nonbond", "bond", "angle", "torsion", "improper") { 
    print_parameters_section($::EMCCharmm::OUT_PARMS, \%items, $_, 0);
  }
  print_cmap($::EMCCharmm::OUT_PARMS);
}


# output topology file

sub print_precedences {
  return if (!scalar(@::EMCCharmm::Precedences));

  my $stream = shift(@_);
  my $n = scalar(@::EMCCharmm::Precedences);

  info("writing $n precedence".($n==1 ? "" : "s")."\n");
  printf($stream "\n# Rule precedences\n\n");
  printf($stream "ITEM\tPRECEDENCE\n\n");
  
  my @arg = ();
  foreach(@::EMCCharmm::Precedences) {
    my @types;
    foreach (split(" ")) { 
      $_ =~ s/[()]//g;
      push(@types, $_);
      push(@arg, $_) if ($_ ne "" && $_ ne "?");
    }
    my $type = @types[0];
    if ($type eq "?") {
      printf($stream "(?\n");
    } elsif ($type eq "") {
      printf($stream ")\n");
    } elsif (scalar(@types)==1) {
      if (defined($::EMCCharmm::TypeDefined{$type})) {
	foreach (sort(keys(%{$::EMCCharmm::TypeDefined{$type}}))) {
	  printf($stream "  (%s)\n", $_);
	}
      }
      if (!defined($::EMCCharmm::TypeCreated{$type})) {
	printf($stream "  (%s)\n", $type);
      }
    } else {
      printf($stream "%s\n", $_);
    }
  }
  check_types(@arg);
  printf($stream "\nITEM\tEND\n");
}


sub print_rules {
  return if (!scalar(@::EMCCharmm::Rules));

  my $id = 0;
  my $stream = shift(@_);
  my $n = scalar(@::EMCCharmm::Rules);

  info("writing $n rule".($n==1 ? "" : "s")."\n");
  printf($stream "\n# Rules\n\n");
  printf($stream "ITEM\tRULES\n\n");
  printf($stream "# id\ttype\telement\tresidue\tatom\tcharge\trule\n\n");
  printf($stream "%s\t%s\n", $id++, 
    join("\t", $::EMCCharmm::Anything, "?", "UNK", "UNK", 0, "*"));
  foreach(sort(@::EMCCharmm::Rules)) {
    my @arg = split("\t");
    @arg[4] = (split("_", @arg[4]))[0];
    check_types(@arg[1]);
    printf($stream "%s\t%s\n", $id++, join("\t", @arg[1,0,4,5,3,2]));
  }
  printf($stream "\nITEM\tEND\n");
}


sub print_templates {
  my $stream = shift(@_);
  my %items = (
    template => {
      header => ["name", "smiles"],
      spec => [\%::EMCCharmm::Templates],
    }
  );

  print_parameters_section($stream, \%items, "template", 0);

  return;
}


sub print_topologies {
  print_header($::EMCCharmm::OUT_TOPOL, "topology", 0);
  print_precedences($::EMCCharmm::OUT_TOPOL);
  print_rules($::EMCCharmm::OUT_TOPOL);
  print_templates($::EMCCharmm::OUT_TOPOL);
}

sub interrupt {
  close_all();
  exit(-1);
}

$SIG{'INT'} = 'interrupt';

# main

{
  init(@ARGV);

  open_all();
  get_header();

  get_masses();
  get_parameters();
  get_define();

  create_residue_index();
  
  analyze();

  my $skip = 0;
  my $trial = $::EMCCharmm::FFTrial;

  if ($trial ne "" && !defined($::EMCCharmm::Residue{index}->{$trial})) {
    close_all();
    error("trial $trial not defined.\n");
  }

  my $continue = 1;

  create_replicas(0);
  $continue = create_increments() if ($::EMCCharmm::Flag{increment});
  reduce_types() if ($::EMCCharmm::Flag{reduce});
  
  if ($continue) {
    get_rules();

    create_replicas(1);
    set_missing(); 
    set_equivalences();
    set_extras();
    set_precedences();

    print_parameters();
    print_topologies();
  }
  close_all();
  message("\n");
}

