#!/usr/bin/env perl
#
#  program:	emc_collapse.pl
#  author:	Pieter J. in 't Veld
#  date:	January 1, 2023.
#  purpose:	Harvest EMC distribtions and combine EMC fields
#
#  Copyright (c) 2004-2023 Pieter J. in 't Veld
#  Distributed under GNU Public License as stated in LICENSE file in EMC root
#  directory
#
#  notes:
#    20230101	Conception date
#

# packages

use File::Basename;

#use lib dirname($0)."/modules";
use lib $ENV{EMC_ROOT}."/scripts/modules";
use EMC;

use strict;


# registration

sub construct {
  my $root = EMC::Common::hash(shift(@_));
  my $global = EMC::Common::hash($root, "global");
  my $field = EMC::Field::construct($root->{field}, {indicator => 0});
  my $message = EMC::Message::construct($root->{message}, {indicator => 0});

  $root->{field} = $field;
  $root->{message} = $message;

  $field->{define}->{angle} = "error";
  $field->{define}->{torsion} = "error";

  set_functions($global);
  set_identity($global);
  set_commands($global);
  set_defaults($global);

  EMC::Options::transcribe($root);

  return $root;
}


# initialization

sub set_identity {
  my $global = EMC::Common::hash(shift(@_));
  
  $global = EMC::Common::attributes(
    EMC::Common::hash($global),
    {
      identity		=> {
	main		=> "EMC",
	script		=> basename($0),
	name		=> "Combine",
	version		=> "1.0",
	date		=> "January 1, 2023",
	author		=> "Pieter J. in 't Veld",
	copyright	=> "2004-".EMC::Common::date_year(),
	command_line	=> "[-option[=#]] target field [...]"
      }
    }
  );
  return $global;
}


sub set_defaults {
  my $global = EMC::Common::hash(shift(@_));

  $global->{flag} = EMC::Common::attributes(
    EMC::Common::hash($global, "flag"),
    {
      compress		=> 0,
      digits		=> 4,
      nfilter		=> 10,
      output		=> "-",
      source		=> ".",
      target		=> ".",
      verbose		=> 0
    }
  );
  return $global;
}


sub set_commands {
  my $global = EMC::Common::hash(shift(@_));
  my $flag = EMC::Common::hash($global, "flag");

  init_output($global);
  $global->{commands} = EMC::Common::attributes(
    EMC::Common::hash($global, "commands"),
    {
      compress		=> {
	comment		=> "compress output",
	set		=> \&set_options,
	default		=> EMC::Math::boolean($flag->{compress})
      },
      digits		=> {
	comment		=> "Number of relevant digits",
	set		=> \&set_options,
	default		=> $flag->{digits}
      },
      help		=> {
	comment		=> "this message",
	set		=> \&EMC::Options::set_help,
      },
      ignore		=> {
	comment		=> "ignore incorrect options",
	set		=> \&set_options,
	default		=> $flag->{ignore}
      },
      output		=> {
	comment		=> "set alternate output file",
	set		=> \&set_options,
	default		=> $flag->{output}
      },
      quiet		=> {
	comment		=> "quiet output",
	set		=> \&set_quiet,
	default		=> EMC::Math::boolean(0)
      },
      source		=> {
	comment		=> "set source directory",
	set		=> \&set_options,
	default		=> $flag->{source}
      },
      target		=> {
	comment		=> "set target directory",
	set		=> \&set_options,
	default		=> $flag->{target}
      },
      verbose		=> {
	comment		=> "set verbose output",
	set		=> \&set_options,
	default		=> EMC::Math::boolean($flag->{verbose})
      }
    }
  );
  $global->{notes} = [
    "A minimum of two fields need to be provided",
    "'-' as output denotes STDOUT"
  ];
  return $global;
}


sub set_options {
  my $struct = shift(@_);
  my $option = EMC::Common::element($struct, "option");
  my $args = EMC::Common::element($struct, "args");
  my $file = EMC::Common::element($struct, "file");
  my $line = EMC::Common::element($struct, "line");
  my $root = EMC::Common::element($struct, "root");
  my $global = EMC::Common::hash($root, "global");
  my $flag = EMC::Common::hash($global, "flag");

  if ($option eq "compress") {
    return $flag->{compress} = EMC::Math::flag($args->[0]);
  } elsif ($option eq "digits") {
    return $flag->{digits} = EMC::Math::eval($args->[0])->[0];
  } elsif ($option eq "ignore") {
    return $flag->{ignore} = EMC::Math::flag($args->[0]);
  } elsif ($option eq "nfilter") {
    return $flag->{nfilter} = EMC::Math::eval($args->[0])->[0];
  } elsif ($option eq "output") {
    return $flag->{output} = $args->[0];
  } elsif ($option eq "source") {
    return $flag->{source} = EMC::IO::scrub_dir($args->[0]);
  } elsif ($option eq "target") {
    return $flag->{target} = EMC::IO::scrub_dir($args->[0]);
  } elsif ($option eq "verbose") {
    return $flag->{verbose} = EMC::Math::flag($args->[0]);
  } else {
    foreach ($root->{set}->{options}) {
      my $result = $_->($struct);
      return $result if (defined($result));
    }
  }
  return undef;
}


sub set_quiet {
  my $struct = shift(@_);
  my $args = EMC::Common::element($struct, "args");
  my $root = EMC::Common::element($struct, "root");

  EMC::Message::set_options({
    option => "message_quiet", root => $root, args => $args});
}


sub set_functions {
  my $global = EMC::Common::hash(shift(@_));
  my $attr = EMC::Common::attributes(@_);
  my $set = EMC::Common::hash($global, "set");
  my $flags = {"commands" => 1};

  $set->{commands} = \&set_commands;
  $set->{defaults} = \&set_defaults;
  $set->{options} = \&set_options;
  foreach (keys(%{$flags})) {
    $set->{flag}->{$_} =
      EMC::Math::flag(defined($attr->{$_}) ? $attr->{$_} : $flags->{$_});
  }
  return $global;
}


sub initialize {
  my $root = construct({files => []});
  my $struct = {root => $root};
  my $global = $root->{global};
  my $flag = $global->{flag};
  my $result = undef;
  my $project = 0;
  my $exist = {};

  foreach (@ARGV) {
    if (substr($_,0,1) eq "-") {
      my @arg = split("=", substr($_,1));
      $struct->{option} = shift(@arg);
      $struct->{args} = [@arg];
      $result = EMC::Options::set_options($root->{options}, $struct);
      if (defined($result) ? 0 : $flag->{ignore} ? 0 : 1) {
	EMC::Options::set_help($struct);
      }
    } else {
      $project = 1;
    }
  }
  init_output($root);
  EMC::Options::set_help($struct) if (!$project);
  set_quiet($root) if ($flag->{output} eq "-");
  EMC::Options::header($root->{options});
  foreach (@ARGV) {
    if (substr($_,0,1) ne "-") {
      my $name = basename(basename($_, ".prm"), ".top");
      if (-e "$name.prm") {
	if (!defined($exist->{$name})) {
	  push(@{$root->{files}}, $name); 
	  $exist->{$name} = 1;
	} else {
	  EMC::Message::warning("omitting reoccuring '$_'\n");
	}
      } else {
	EMC::Message::warning("'$_' not found\n");
      }
    }
  }
  if (!scalar(@{$root->{files}})) {
    EMC::Options::header($root->{options});
    EMC::Message::error("no files found.\n");
  }
  return $root;
}


sub init_output {
  my $root = shift(@_);
  my $global = EMC::Common::hash($root, "global");
  my $flag = EMC::Common::hash($global, "flag");

  if (substr($flag->{output}, -3) eq ".gz") {
    $flag->{output} = substr($flag->{output}, 0, length($flag->{output})-3);
    $flag->{compress} = 1;
  }
  if ($flag->{output} ne "-") {
    $flag->{output} .= ".redefine" if (substr($flag->{output}, -9) ne ".redefine");
    $flag->{output} .= ".gz" if ($flag->{compress});
  }
}


# functions

sub get_index {
  my $index = shift(@_);
  my $key = shift(@_);
  my $i = 0;

  foreach (@{$index}) {
    last if ($_ eq $key);
    ++$i;
  }
  return $i<scalar(@{$index}) ? $i : -1;
}


# processing

# i/o

sub field_import {
  my $name = shift(@_);
  my $attr = EMC::Common::attributes(shift(@_));
  my $ext = EMC::Common::element($attr, "ext");
  my $field = EMC::Common::element($attr, "field");
  my $stream;

  $ext = [".prm"] if (!defined($ext));
  foreach (@{$ext}) {
    EMC::Message::info("reading field from '$name$_'\n");
    $stream = EMC::IO::open($name.$_, "r");
    $field = EMC::Field::read($stream, {field => $field});
    EMC::IO::close($stream, $name);
  }
  rules_frequency($field);
  mass_frequency($field);
  return $field;
}


sub field_export {
  my $name = shift(@_);
  my $field = shift(@_);
  my $stream;

  EMC::Message::info("writing field to '$name'\n");
  $stream = EMC::IO::open($name, "w");
  $field = EMC::Field::write($stream, $field);
  EMC::IO::close($stream, $name);
  
  return $field;
}


sub print_initial {
  my $text = scalar(@_)>1 ? shift(@_) : "";
  my $initial = shift(@_);
  my $first = 1;

  EMC::Message::info($text."{\n");
  foreach (sort(keys(%{$initial}))) {
    my $ptr = $initial->{$_};

    print($first ? "" : ",\n");
    printf("\t%s\t=> %s", $_, 
      ref($ptr) eq "HASH" ? "{".join(", ", sort(keys(%{$initial->{$_}})))."}" :
      ref($ptr) eq "ARRAY" ? "{".join(", ", sort(@{$ptr}))."}" : $ptr
    );
    $first = 0;
  }
  print("}\n");
}


sub print_output {
  my $root = shift(@_);
  my $convert = shift(@_);
  my $name = EMC::Common::element($root, "global", "flag", "output");

  $name = "-" if (!defined($name));
  
  my $identity = EMC::Common::element($root, "global", "identity");
  my $date = EMC::Common::date_full();
  my $stream = EMC::IO::open($name, "w");

  EMC::Message::info("writing redefinitions to '$name'\n");
  printf($stream "#
#  CHARMM redefinition file to be used for consolidation of force field files
#  converted by $identity->{script} $identity->{version}, $identity->{date} by $identity->{author},
#  created on $date.
#

ITEM	REDEFINE

# src	type	pair	incr	bond	angle	torsion	impr

");

  foreach (sort(keys(%{$convert}))) {
    printf($stream "%s", $_);
    for (my $i=0; $i<7; ++$i) {
      printf($stream "\t%s", $convert->{$_});
    }
    print($stream "\n");
  }
  printf($stream "\nITEM\tEND\n");
  EMC::IO::close($stream, $name);
}


# processing

sub process {
  my $root = shift(@_);
  my $fields = {};

  foreach (@{$root->{files}}) {
    my $name = basename(basename($_, ".prm"), ".top");
    $fields->{$name} = field_import(
      $name, {ext => [".prm", ".top"], field => {}});
    #EMC::Message::dumper("frequency = ", $fields->{$name}->{mass}->{frequency});
  };
  return $root->{fields} = $fields;
}


# adaptations

sub rules_frequency {
  my $field = shift(@_);
  my $rules = EMC::Common::element($field, "rules");
  my $data = EMC::Common::element($field, "rules", "data");
  my $frequency = {};

  return if (!defined($data));

  foreach (@{$data}) {
    ++$frequency->{$_->[1]};
  }
}


sub mass_frequency {
  my $field = shift(@_);
  my $mass = EMC::Common::element($field, "mass");
  my $data = EMC::Common::element($field, "mass", "data");
  my $ielement = get_index($mass->{index}, "element");
  my $ifreq = get_index($mass->{index}, "freq");
  my $frequency = {};
  my $element = {};

  return if (!defined($data) || $ifreq<0 || $ielement<0);

  $ifreq -= $mass->{flag}->{ntypes};
  $ielement -= $mass->{flag}->{ntypes};
  foreach (keys(%{$data})) {
    my $ptr = $data->{$_};
    my $freq = eval($ptr->[$ifreq]);

    $frequency->{$_} = $freq if ($freq);
    $element->{$_} = $ptr->[$ielement] if ($freq);
  }
  $mass->{frequency} = $frequency;
  $mass->{element} = $element;
}


# inversion

sub invert {
  my $root = shift(@_);
  my $fields = shift(@_);
  my $allowed = 
    defined(@_[0]) ? shift(@_) : ["nonbond", "bond", "angle", "torsion"];

  my $digits = EMC::Common::element($root, "global", "flag", "digits");
  my $inverts = {};

  EMC::Message::info("creating inversions\n");
  
  foreach (keys(%{$fields})) {
    my $field = $fields->{$_};
    my $invert = $inverts;
    
    foreach (defined($allowed) ? @{$allowed} : sort(keys(%{$field}))) {
      next if (!defined($field->{$_}));
      $invert->{$_} = {} if (!defined($invert->{$_}));

      my $sub = $invert->{$_};
      my $flag = $field->{$_}->{flag};
      my $ntypes = $flag->{ntypes};
      my $data = $field->{$_}->{data};
      my $keys = [[sort(keys(%{$data}))]];
      my $n = [scalar(@{$keys->[0]})];
      my $ptr = [$data];
      my $types = [];
      my $i = [];
      my $j = 0;

      for (my $k=0; $k<$ntypes; ++$k) {
       	$i->[$k] = 0;
      }
      while ($j>=0) {
	$types->[0] = $keys->[0]->[$i->[0]];		# init arrays
	for (my $k=1; $k<$ntypes; ++$k) {
	  $ptr->[$k] = $ptr->[$k-1]->{$keys->[$k-1]->[$i->[$k-1]]};
	  $keys->[$k] = [sort(keys(%{$ptr->[$k]}))];
	  $n->[$k] = scalar(@{$keys->[$k]});
	  $types->[$k] = $keys->[$k]->[$i->[$k]];
	}

	my $key = join(":", 
	  map(
	    {EMC::Math::digits($_, $digits)} 
	    @{$ptr->[-1]->{$keys->[-1]->[$i->[-1]]}})
	);
	$sub->{$key} = [] if (!defined($sub->{$key}));
	push(@{$sub->{$key}}, [@{$types}]);
	
	for ($j=$ntypes-1; $j>=0; --$j) {		# increase counter
	  last if (++$i->[$j]<$n->[$j]);
	  $i->[$j] = 0;
	}
      }
    }
  }
  return $root->{inverts} = $inverts;
}


sub invert_purge {
  my $inverts = shift(@_);
  my $fields = shift(@_);

  EMC::Message::info("purging inversions\n");
  
  foreach (keys(%{$inverts})) {
    my $invert = $inverts->{$_};

    foreach (keys(%{$invert})) {
      my $list = $invert->{$_};
      my $flag = {};

      delete($invert->{$_}) if (scalar(@{$list})<2);
      foreach (@{$list}) {
	foreach (@{$_}) {
	  my $type = $_;
	  foreach (keys(%{$fields})) {
	    if (defined($fields->{$_}->{mass}->{data}->{$type})) {
	      $flag->{$_} = 1;
	    }
	  }
	}
      }
      delete($invert->{$_}) if (scalar(keys(%{$flag}))<2);
    }
  }
  return $inverts;
}


# conversion

sub separate {
  my $type = shift(@_);
  my $mass = shift(@_);
  my $collect = {target => {}, other => {}};
  my @ptr = map({$collect->{$_}} ("target", "other"));

  foreach (@{@_[0]}) {
    foreach (@{$_}) {
      next if (defined($type->{$_}) && !defined($mass->{$_}));
      @ptr[defined($type->{$_}) ? 0 : 1]->{$_} = 1;
    }
  }
  return $collect;
}


sub members {
  my $inverts = shift(@_);
  my $members = {};

  foreach (keys(%{$inverts})) {
    my $invert = $inverts->{$_};
    my $member = defined($members->{$_}) ? $members->{$_} : $members->{$_} = {};

    foreach (keys(%{$invert})) {
      foreach (@{$invert->{$_}}) {
	foreach (@{$_}) {
	  $member->{$_} = 1;
	}
      }
    }
  }
  return $members;
}


sub filter {
  my $candidates = shift(@_);
  my $type = shift(@_);
  my $mass = shift(@_);
  my $inverts = shift(@_);
  my $invert = EMC::Common::element($inverts, shift(@_));
  my $debug = shift(@_);

  foreach (sort(keys(%{$invert}))) {
    my $collect = separate($type, $mass, $invert->{$_});
    EMC::Message::spot("key = $_\n\n") if ($debug);
    foreach (sort(keys(%{$collect->{other}}))) {
      next if (!defined($candidates->{$_}));
      my $ptr = $candidates->{$_};
      my $tmp = {%{$ptr}};
      EMC::Message::keys("key[$_] = ", $ptr) if ($debug);
      foreach (keys(%{$collect->{target}})) {
	delete($tmp->{$_}) if (defined($tmp->{$_}));
      }
      foreach (keys(%{$tmp})) {
	delete($ptr->{$_});
      }
    }
    print("\n") if ($debug);
  }
}


sub transfer {
  my $candidates = shift(@_);
  my $convert = shift(@_);
  my $initial = shift(@_);
  my $used = shift(@_);
  my $tmp;

  foreach (keys(%{$candidates})) {
    next if (keys(%{$tmp = $candidates->{$_}})!=1);
    $used->{$convert->{$_} = (keys(%{$tmp}))[0]} = 1;
    delete($initial->{$_});
  }
  foreach (keys(%{$initial})) {
    foreach (keys(%{$tmp = $initial->{$_}})) {
      delete($tmp->{$_}) if (defined($used->{$_}));
    }
    next if (($tmp = scalar(keys(%{$initial->{$_}})))>1);
    $used->{$convert->{$_} = (keys(%{$initial->{$_}}))[0]} = 1 if ($tmp==1);
    delete($initial->{$_});
  }
}


sub convert {
  my $inverts = shift(@_);
  my $fields = shift(@_);
  my $target = shift(@_);

  my $type = EMC::Common::element($fields, $target, "equivalence", "data");
  my $mass = EMC::Common::element($fields, $target, "mass", "data");
  my $invert = EMC::Common::element($inverts, "nonbond");
  my $members = members($inverts);
  my $candidates = {};
  my $initial = {};
  my $convert = {};
  my $used = {};
  my $debug = 0;
  my ($collect, $ptr, $tmp, $n);

  # initialization

  EMC::Message::info("initialize conversion\n");
  foreach (keys(%{$invert})) {				# nonbonds
    $collect = separate($type, $mass, $invert->{$_});
    foreach (sort(keys(%{$collect->{other}}))) {
      $initial->{$_} = {%{$collect->{target}}};
    }
  }

  # pass 1: identify universal types

  my $candidates = EMC::Element::deep_copy($initial);

  EMC::Message::info("identify universal types\n");
  print_initial("candidates = ", $candidates);
  print_initial($initial) if ($debug);
  foreach ("bond", "angle", "torsion") {
    filter($candidates, $type, $mass, $inverts, $_);
  }
  transfer($candidates, $convert, $initial, $used);
  
  # pass 2: determine 
  
  EMC::Message::info("identify specific types\n");
  print_initial($initial) if ($debug);
  foreach ("bond", "angle", "torsion") {
    $candidates = EMC::Element::deep_copy($initial);
    filter($candidates, $type, $mass, $inverts, $_);
    transfer($candidates, $convert, $initial, $used);
  }

  # end

  foreach (keys(%{$convert})) {
    $convert->{$_} = $_ if (!defined($convert->{$_}));
  }
  print_initial($initial) if ($debug);
  return $convert;
}


# validate cross referenced types

sub validate {
  my $root = shift(@_);
  my $convert = shift(@_);
  my $field = shift(@_);
  my $validated = shift(@_);

  my $debug = 1;
  my $digits = EMC::Common::element($root, "global", "flag", "digits");

  # check for clashes in constants as a result of crosslinked types
  
  EMC::Message::info("validating conversions\n");
  $validated = {} if (!defined($validated));
  foreach ("nonbond", "bond", "angle", "torsion", "improper")
  {
    my $from = EMC::Common::element($field, $_);

    next if (!defined($from));

    my $to = EMC::Common::hash($validated, $_);
    my $ntypes = $from->{flag}->{ntypes};
    my $data = $from->{data};
    my $keys = [[sort(keys(%{$data}))]];
    my $n = [scalar(@{$keys->[0]})];
    my $ptr = [$data];
    my $types = [];
    my $i = [];
    my $j = 0;

    while ($j>=0) {
      $types->[0] = $keys->[0]->[$i->[0]];		# init arrays
      for (my $k=1; $k<$ntypes; ++$k) {
	$ptr->[$k] = $ptr->[$k-1]->{$keys->[$k-1]->[$i->[$k-1]]};
	$keys->[$k] = [sort(keys(%{$ptr->[$k]}))];
	$n->[$k] = scalar(@{$keys->[$k]});
	$types->[$k] = $keys->[$k]->[$i->[$k]];
      }

      my $p = $to;					# compare
      my @d = map(
	  {EMC::Math::digits($_, $digits)} 
	  @{$ptr->[-1]->{$keys->[-1]->[$i->[-1]]}}
      );
      my @t = map({defined($convert->{$_}) ? $convert->{$_} : $_} @{$types});
      
      for (my $i=0; $i<$ntypes-1; ++$i) {
	$p = EMC::Common::hash($p, @t[$i]);
      }
      $p = EMC::Common::array($p, @t[-1]);
      if (!scalar(@{$p})) {
	@{$p} = @d;
      } elsif(EMC::List::compare(@{$p}, @d)) {
	EMC::Message::warning("\t",
	  join("\t", "INCOMP", $_, @t, @{$types}), "\n");
      }
      
      for ($j=$ntypes-1; $j>=0; --$j) {			# increase counter
	last if (++$i->[$j]<$n->[$j]);
	$i->[$j] = 0;
      }
    }
  }
  return $validated;
}


# output

sub print_invert {
  my $inverts = shift(@_);

  foreach (@_) {
    my $sub = $_;

    foreach (sort(keys(%{$inverts}))) {
      my $invert = $inverts->{$_};

      next if (defined($sub) && $sub ne $_);
      print("\nfield = $_\n\n");
      foreach (sort(keys(%{$invert}))) {
	my $key = $_;
	my $list = $invert->{$_};
	my $s = [];

	foreach (@{$list}) {
	  push(@{$s}, join(":", @{$_}));
	}
	printf("%-30.30s %s\n", $key, join(", ", @{$s}));
      }
      print("\n");
    }
  }
}


# global

{
  my $root = initialize(@ARGV);
  my $fields = process($root);
  my $files = $root->{files};
  my $target = $files->[0];
  my $inverts = invert($root,
    EMC::Hash::select($fields, $files->[0], $files->[1]));
  my $verbose = EMC::Common::element($root, "global", "flag", "verbose");
  
  print_invert($inverts, "nonbond") if ($verbose);
  invert_purge($inverts, $fields);
  print_invert($inverts, "nonbond", "bond", "angle", "torsion") if ($verbose);
  
  my $convert = convert($inverts, $fields, $target);
  my $validated = validate($root, $convert, $fields->{$target});
  
  print_initial("conversion = ", $convert);
  $validated = validate($root, $convert, $fields->{$files->[1]}, $validated);
  
  EMC::Message::dumper("validated = ", $validated) if ($verbose);

  print_output($root, $convert);
  print("\n");
}
